import multiprocessing
from src.rlpe.Agents.RL_agents.q_learning_agents import KERAS_DQN
from src.rlpe.Transforms.single_taxi_transforms import *
from src.rlpe.constants import *


def check_success_rate():
    success_rates = {}
    for file in os.listdir(TRAINED_AGENT_ON_SPECIFIC_TRANSFORM_NUM_PATH):
        success_rate = visualize_trained_agent(file)
        success_rates[file[:-4]] = success_rate

    for transform_name, success_rate in success_rates.items():
        if success_rate > 0.6:
            a = 7


def visualize_trained_agent(file_name):
    transform_name = file_name[:-4] if file_name[-4:] == ".pkl" else file_name
    file = open(TRAINED_AGENT_ON_SPECIFIC_TRANSFORM_NUM_PATH + file_name, "rb")
    new_env = pickle.load(file)

    agent_name = KERAS_DQN
    model_name = agent_name + '_' + transform_name
    dir_path = TRAINED_AGENTS_DIR_PATH + model_name + '/'
    new_agent = rl_agent.create_agent(new_env, agent_name)  # TODO - to fix this two lines
    new_agent = load_existing_agent(new_env, agent_name, transform_name, dir_path + model_name)
    evaluation_result = rl_agent.run(new_agent, 5, method=EVALUATE)

    result_name = transform_name + "_" + agent_name
    result_dir = result_name + "_200000/"
    file = open(TRAINED_AGENT_RESULTS_PATH + result_dir + result_name + "_explanation.pkl", "rb")
    explanation = pickle.load(file)
    file = open(TRAINED_AGENT_RESULTS_PATH + result_dir + result_name + "_result.pkl", "rb")
    result = pickle.load(file)

    return result[transform_name][SUCCESS_RATE]


def create_single_taxi_transforms():
    cur_env_preconditions = load_pkl_file(PRECONDITIONS_PATH)
    processes = []
    functions = [generate_single_transforms, generate_double_of_transforms, generate_triple_of_transforms]
    for func in functions:
        p = multiprocessing.Process(target=func, args=(cur_env_preconditions,))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()


from src.rlpe.experiments import *
from src.rlpe.save_load_utils import *
from src.rlpe.constants import *
from src.rlpe.save_load_utils import load_transform_by_name


def generate_agent(original_env, agent_name, num_of_episodes, transformed_env, transform_name):
    result = dict()
    anticipated_policy = ANTICIPATED_POLICY
    explanation = []

    result, explanation = create_run_and_evaluate_agent(original_env, transformed_env, agent_name, transform_name,
                                                        num_of_episodes, anticipated_policy, result, explanation)
    satisfaction = not (explanation is None or len(explanation) == 0)
    if not satisfaction:
        print(f"{transform_name} is not your answer! it is not an explanation")
    else:
        print(f"explanation found {explanation}!!")

    dir_name = TRAINED_AGENT_RESULTS_PATH + transform_name + "_" + agent_name + "_200000"
    make_dir(dir_name)
    save_pkl_file(dir_name + "/" + transform_name + "_" + agent_name + "_result", result)
    save_pkl_file(dir_name + "/" + transform_name + "_" + agent_name + "_explanation", explanation)
    return satisfaction


# if __name__ == '__main__':
#     path = sys.argv[1:][0]
#     file_name = os.path.basename(path)
#     transform_name, new_env = load_transform_by_name(TRAINED_AGENT_SPECIFIC_PATH + file_name)
#     a = generate_agent(SINGLE_TAXI_EXAMPLE, KERAS_DQN, ITER_NUM, new_env, transform_name)



from src.rlpe.Environments.LunarLander.lunar_lander import *
from src.rlpe.Transforms.transform_constants import *


class LunarLanderTransformedEnv(LunarLander):
    def __init__(self, transforms):
        self.in_reset = True
        super().__init__()
        self.in_reset = False

        x_coordinate, y_coordinate = transforms[0], transforms[1]
        x_horizontal_velocity, y_vertical_velocity = transforms[2], transforms[3]
        orientation_in_space = transforms[4]
        angular_velocity = transforms[5]

        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.x_horizontal_velocity = x_horizontal_velocity
        self.y_vertical_velocity = y_vertical_velocity
        self.orientation_in_space = orientation_in_space
        self.angular_velocity = angular_velocity

    def step(self, action):
        if self.continuous:
            action = np.clip(action, -1, +1).astype(np.float32)
        else:
            assert self.action_space.contains(action), "%r (%s) invalid " % (action, type(action))

        # Engines
        tip = (math.sin(self.lander.angle), math.cos(self.lander.angle))
        side = (-tip[1], tip[0])
        dispersion = [self.np_random.uniform(-1.0, +1.0) / SCALE for _ in range(2)]

        m_power = 0.0
        if (self.continuous and action[0] > 0.0) or (not self.continuous and action == 2):
            # Main engine
            if self.continuous:
                m_power = (np.clip(action[0], 0.0, 1.0) + 1.0) * 0.5  # 0.5..1.0
                assert m_power >= 0.5 and m_power <= 1.0
            else:
                m_power = 1.0
            ox = tip[0] * (4 / SCALE + 2 * dispersion[0]) + side[0] * dispersion[
                1]  # 4 is move a bit downwards, +-2 for randomness
            oy = -tip[1] * (4 / SCALE + 2 * dispersion[0]) - side[1] * dispersion[1]
            impulse_pos = (self.lander.position[0] + ox, self.lander.position[1] + oy)
            # particles are just a decoration, 3.5 is here to make particle speed adequate
            p = self._create_particle(3.5, impulse_pos[0], impulse_pos[1], m_power)
            p.ApplyLinearImpulse((ox * MAIN_ENGINE_POWER * m_power, oy * MAIN_ENGINE_POWER * m_power), impulse_pos,
                                 True)
            self.lander.ApplyLinearImpulse((-ox * MAIN_ENGINE_POWER * m_power, -oy * MAIN_ENGINE_POWER * m_power),
                                           impulse_pos, True)

        s_power = 0.0
        if (self.continuous and np.abs(action[1]) > 0.5) or (not self.continuous and action in [1, 3]):
            # Orientation engines
            if self.continuous:
                direction = np.sign(action[1])
                s_power = np.clip(np.abs(action[1]), 0.5, 1.0)
                assert 0.5 <= s_power <= 1.0
            else:
                direction = action - 2
                s_power = 1.0
            ox = tip[0] * dispersion[0] + side[0] * (3 * dispersion[1] + direction * SIDE_ENGINE_AWAY / SCALE)
            oy = -tip[1] * dispersion[0] - side[1] * (3 * dispersion[1] + direction * SIDE_ENGINE_AWAY / SCALE)
            impulse_pos = (self.lander.position[0] + ox - tip[0] * 17 / SCALE,
                           self.lander.position[1] + oy + tip[1] * SIDE_ENGINE_HEIGHT / SCALE)
            p = self._create_particle(0.7, impulse_pos[0], impulse_pos[1], s_power)
            p.ApplyLinearImpulse((ox * SIDE_ENGINE_POWER * s_power, oy * SIDE_ENGINE_POWER * s_power), impulse_pos,
                                 True)
            self.lander.ApplyLinearImpulse((-ox * SIDE_ENGINE_POWER * s_power, -oy * SIDE_ENGINE_POWER * s_power),
                                           impulse_pos, True)

        self.world.Step(1.0 / FPS, 6 * 30, 2 * 30)

        pos = self.lander.position
        vel = self.lander.linearVelocity
        state = [
            (pos.x - VIEWPORT_W / SCALE / 2) / (VIEWPORT_W / SCALE / 2),
            (pos.y - (self.helipad_y + LEG_DOWN / SCALE)) / (VIEWPORT_H / SCALE / 2),
            vel.x * (VIEWPORT_W / SCALE / 2) / FPS,
            vel.y * (VIEWPORT_H / SCALE / 2) / FPS,
            self.lander.angle,
            20.0 * self.lander.angularVelocity / FPS,
            1.0 if self.legs[0].ground_contact else 0.0,
            1.0 if self.legs[1].ground_contact else 0.0
        ]
        assert len(state) == 8

        reward = 0
        shaping = \
            - 100 * np.sqrt(state[0] * state[0] + state[1] * state[1]) \
            - 100 * np.sqrt(state[2] * state[2] + state[3] * state[3]) \
            - 100 * abs(state[4]) + 10 * state[6] + 10 * state[7]  # And ten points for legs contact, the idea is if you
        # lose contact again after landing, you get negative reward
        if self.prev_shaping is not None:
            reward = shaping - self.prev_shaping
        self.prev_shaping = shaping

        reward -= m_power * 0.30  # less fuel spent is better, about -30 for heurisic landing
        reward -= s_power * 0.03

        done = False
        if self.game_over or abs(state[0]) >= 1.0:
            done = True
            reward = -100
        if not self.lander.awake:
            done = True
            reward = +100
        if not self.in_reset:
            if self.x_coordinate:
                state[0] = 0.0
            if self.y_coordinate:
                state[1] = 0.0
            if self.x_horizontal_velocity:
                state[2] = 0.0
            if self.y_vertical_velocity:
                state[3] = 0.0
            if self.orientation_in_space:
                state[4] = 0.0
            if self.angular_velocity:
                state[5] = 0.0
        return np.array(state, dtype=np.float32), reward, done, {}


def get_lunar_lander_transform_name(transforms):
    x_coordinate, y_coordinate = transforms[0], transforms[1]
    x_horizontal_velocity, y_vertical_velocity = transforms[2], transforms[3]
    orientation_in_space = transforms[4]
    angular_velocity = transforms[5]
    name = ""
    name += X_COORDINATE if x_coordinate else ""
    name += Y_COORDINATE if y_coordinate else ""
    name += X_VELOCITY if x_horizontal_velocity else ""
    name += Y_VELOCITY if y_vertical_velocity else ""
    name += ORIENTATION if orientation_in_space else ""
    name += ANGULAR_VELOCITY if angular_velocity else ""
    return name


if __name__ == '__main__':
    # demo_heuristic_lander(LunarLander(), render=True)
    transforms = [True, True, True, True, True, True]
    demo_heuristic_lander(LunarLanderTransformedEnv(transforms), render=True)



from abc import ABC, abstractmethod


class AbstractExpert(ABC):
    @abstractmethod
    def get_expert_policy_set(self, state_tuple):
        pass

    @abstractmethod
    def full_expert_policy_dict(self):
        pass



RLLIB_A2C = "rllib_a2c"
RLLIB_A3C = "rllib_a3c"
RLLIB_BC = "rllib_bc"
RLLIB_DQN = "rllib_dqn"
RLLIB_APEX_DQN = "rllib_apex_dqn"
RLLIB_IMPALA = "rllib_impala"
RLLIB_MARWIL = "rllib_marwil"
RLLIB_PG = "rllib_pg"
RLLIB_PPO = "rllib_ppo"
RLLIB_APPO = "rllib_appo"
RLLIB_SAC = "rllib_sac"
RLLIB_LIN_UCB = "rllib_lin_usb"
RLLIB_LIN_TS = "rllib_lin_ts"

NUM_WORKERS = 1
WITH_DEBUG = True
NUM_GPUS = 0
g_config = {}

TAXI = "taxi"
SPEAKER_LISTENER = "simple_speaker_listener"

agents_gamma = {'taxi_1': 0.85, 'taxi_2': 0.95, 'taxi_3': 0.85, 'taxi_4': 0.95}


def is_rllib_agent(agent_name):
    return agent_name == RLLIB_A2C or \
           agent_name == RLLIB_A3C or \
           agent_name == RLLIB_BC or \
           agent_name == RLLIB_DQN or \
           agent_name == RLLIB_APEX_DQN or \
           agent_name == RLLIB_IMPALA or \
           agent_name == RLLIB_MARWIL or \
           agent_name == RLLIB_PG or \
           agent_name == RLLIB_PPO or \
           agent_name == RLLIB_APPO or \
           agent_name == RLLIB_SAC or \
           agent_name == RLLIB_LIN_UCB or \
           agent_name == RLLIB_LIN_TS


def get_rllib_agent(agent_name, env_name, env, env_to_agent):
    config = get_config(env_name, env, 1) if is_rllib_agent(agent_name) else {}
    if agent_name == RLLIB_A2C:
        import ray.rllib.agents.a3c as a2c
        agent = a2c.A2CTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_A3C:
        import ray.rllib.agents.a3c as a3c
        agent = a3c.A3CTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_BC:
        import ray.rllib.agents.marwil as bc
        agent = bc.BCTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_DQN:
        import ray.rllib.agents.dqn as dqn
        agent = dqn.DQNTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_APEX_DQN:
        import ray.rllib.agents.dqn as dqn
        agent = dqn.ApexTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_IMPALA:
        import ray.rllib.agents.impala as impala
        agent = impala.ImpalaTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_MARWIL:
        import ray.rllib.agents.marwil as marwil
        agent = marwil.MARWILTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_PG:
        import ray.rllib.agents.pg as pg
        agent = pg.PGTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_PPO:
        import ray.rllib.agents.ppo as ppo
        agent = ppo.PPOTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_APPO:
        import ray.rllib.agents.ppo as ppo
        agent = ppo.APPOTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_SAC:
        import ray.rllib.agents.sac as sac
        agent = sac.SACTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_LIN_UCB:
        import ray.rllib.contrib.bandits.agents.lin_ucb as lin_ucb
        agent = lin_ucb.LinUCBTrainer(config=config, env=env_to_agent)
    elif agent_name == RLLIB_LIN_TS:
        import ray.rllib.contrib.bandits.agents.lin_ts as lin_ts
        agent = lin_ts.LinTSTrainer(config=config, env=env_to_agent)
    return agent


def get_config(env_name, env, number_of_agents):
    """
    :param env_name
    :param env:
    :param number_of_agents:
    :return:
    """
    config = {}
    if env_name == TAXI:
        if number_of_agents == 1:  # single agent config
            config = {"num_gpus": NUM_GPUS, "num_workers": NUM_WORKERS}
        else:  # multi-agent config
            policies = get_multi_agent_policies(env, number_of_agents)
            config = {'multiagent': {'policies': policies, "policy_mapping_fn": lambda taxi_id: taxi_id},
                      "num_gpus": NUM_GPUS,
                      "num_workers": NUM_WORKERS}
    elif env_name == SPEAKER_LISTENER:
        config = {
            "num_gpus": 0,
            "lr_schedule": [[0, 0.007], [20000000, 0.0000000001]],
            "framework": "torch",
            "env_config": {"name": "simple_speaker_listener"},
            "clip_rewards": True,
            "num_envs_per_worker": 1,
            "rollout_fragment_length": 20,
            "monitor": True,
        }
    global g_config
    g_config = config
    return config


def get_multi_agent_policies(env, number_of_agents):
    policies = {}
    for i in range(number_of_agents):
        name = 'taxi_' + str(i + 1)
        policies[name] = (None, env.observation_space, env.action_space, {'gamma': agents_gamma[name]})
    return policies



import numpy as np

from src.rlpe.Agents.abstract_agent import AbstractAgent


class ValueIterationAgent(AbstractAgent):
    def __init__(self, env):
        super().__init__(env)
        self.nA = len(self.env.get_available_actions_dictionary()[0])
        self.nS = self._get_state_number()
        self.policy_dict = {}

    def run(self):
        self.value_iteration()

    def value_iteration(self, theta=1, discount_factor=0.99, display=True):
        """
        Value Iteration Algorithm.

        Args:
            # env: OpenAI env. env.P represents the transition probabilities of the environment.
            #     env.P[s][a] is a list of transition tuples (prob, next_state, reward, done).
            #     env.nS is a number of states in the environment.
            #     env.nA is a number of actions in the environment.
            theta: We stop evaluation once our value function change is less than theta for all states.
            discount_factor: Gamma discount factor.
            display: displays iteration num

        Returns:
            A tuple (policy, V) of the optimal policy and the optimal value function.
        """

        def one_step_lookahead(state, V):
            """
            Helper function to calculate the value for all action in a given state.

            Args:
                state: The state to consider (int)
                V: The value to use as an estimator, Vector of length env.nS

            Returns:
                A vector of length env.nA containing the expected value of each action.
            """
            A = np.zeros(self.nA)
            state = list(self.env.decode(state))
            self.env.reset()
            self.env.state = state
            for a in range(self.nA):
                taxi_name = "taxi_1"
                next_state, reward, done, _ = self.env.step({taxi_name: a})
                next_state, reward, done, prob = next_state[taxi_name][0], reward[taxi_name], done[taxi_name], 1.0
                # for prob, next_state, reward, done in env.P[state][a]:
                cur_next_state = self.env.encode(next_state)
                #  taxi_row, taxi_col, fuel, pass_loc_x, pass_loc_y, dest_idx_x, dest_idx_y, pass_status
                A[a] += prob * (reward + discount_factor * V[cur_next_state - 1])
            return A

        V = np.zeros(self.nS)
        i = 0
        while True:
            # Stopping condition
            delta = 0
            # Update each state...
            for s in range(self.nS):
                # Do a one-step lookahead to find the best action
                A = one_step_lookahead(s, V)
                best_action_value = np.max(A)
                # Calculate delta across all states seen so far
                delta = max(delta, np.abs(best_action_value - V[s]))
                # Update the value function. Ref: Sutton book eq. 4.10.
                V[s] = best_action_value
                # Check if we can stop
            i += 1
            if display:
                print("Optimal agent iteration num: ", i)
            if delta < theta:
                break

        # Create a deterministic policy using the optimal value function
        policy = np.zeros([self.nS, self.nA])  # TODO - policies = set of policies
        policy_dict = {}
        for s in range(self.nS):
            decoded_state = list(self.env.decode(s))
            flatten_state = decoded_state[0][0] + decoded_state[1] + decoded_state[2][0] + decoded_state[3][0] + \
                            decoded_state[4]
            # One step lookahead to find the best action for this state
            A = one_step_lookahead(s, V)
            best_action = np.argmax(A)  # TODO - take all the actions that maximize the value / threshold
            # Always take the best action
            policy[s, best_action] = 1.0
            policy_dict[tuple(flatten_state)] = best_action
        self.policy_dict = policy_dict
        return policy_dict, policy, V

    def _get_state_number(self):
        taxi_possible_locations = self.env.num_rows * self.env.num_columns
        fuel = self.env.max_fuel[0] + 1
        passenger_possible_locations = self.env.num_rows * self.env.num_columns
        passenger_possible_destinations = len(self.env.passengers_locations)
        passengers_status = 3
        return taxi_possible_locations * fuel * passenger_possible_locations * passenger_possible_destinations * passengers_status

    def compute_action(self, state):
        return self.policy_dict[tuple(state[0])]

    def stop_episode(self):
        pass

    def episode_callback(self, state, action, reward, next_state, terminated):
        pass



import copy
from collections import defaultdict

from src.rlpe.Transforms.single_taxi_transforms import generate_transformed_env


def cluster_transforms_by_features(preconditions):
    feature_clusters = defaultdict(dict)  # {action: {idx : [[values]]}}
    for action, action_info in preconditions.items():
        for idx, values in action_info.items():
            for val in values:
                if action in feature_clusters[idx].keys():
                    if idx in feature_clusters[idx][action].keys():
                        feature_clusters[idx][action][idx].append(list(val))
                    else:
                        feature_clusters[idx][action][idx] = [list(val)]
                else:
                    feature_clusters[idx][action] = {idx: [list(val)]}
    return feature_clusters


def calculate_possible_satisfaction_rate(transformed_env, anticipated_policy):
    pos, neg, = 0, 0
    for anticipated_state, anticipated_actions in anticipated_policy.items():
        states_from_anticipated_state = transformed_env.get_states_from_partial_obs(anticipated_state)
        for state in states_from_anticipated_state:
            transformed_env.s = state
            for anticipated_action in anticipated_actions:
                next_state, reward, done, info = transformed_env.step(anticipated_action)
                if next_state != state:
                    pos += 1
                else:
                    neg += 1
    all_res = pos + neg
    return pos / all_res


def refactor_max_cluster(max_cluster):
    max_cluster_list = []
    for act, act_info in max_cluster.items():
        for idx, values in act_info.items():
            for val in values:
                max_cluster_list.append((act, idx, val))
    return max_cluster_list


def intersection(lst1, lst2):
    lst3 = [value for value in lst1 if value in lst2]
    return lst3


def dict_intersection(working_cluster, temp_cluster):
    working_cluster_list = refactor_max_cluster(working_cluster)
    temp_cluster_list = refactor_max_cluster(temp_cluster)
    cluster_intersection = intersection(working_cluster_list, temp_cluster_list)
    new_working_cluster = defaultdict(dict)
    for action, idx, val in cluster_intersection:
        if action in new_working_cluster.keys():
            if idx in new_working_cluster[action].keys():
                new_working_cluster[action][idx].append(list(val))
            else:
                new_working_cluster[action][idx] = [list(val)]
        else:
            new_working_cluster[action] = {idx: [list(val)]}
    return new_working_cluster


def greedy_search(preconditions, anticipated_policy):
    clusters = cluster_transforms_by_features(preconditions.not_allowed_features)
    transformed_envs1, satisfaction_rates1, transformed_envs2, satisfaction_rates2 = [], [], [], []
    max_transformed_env_opt, best_cluster = None, None
    k, msr_opt = 0, 0.0
    while k < 100 and msr_opt != 1.0:
        msr, max_cluster, max_transformed_env = 0, None, None
        for cluster in clusters.values():
            transformed_env = generate_transformed_env(cluster, "", save=False, try_load_existing=False)
            satisfaction_rate = calculate_possible_satisfaction_rate(transformed_env, anticipated_policy)
            satisfaction_rates1.append(satisfaction_rate)
            transformed_envs1.append(transformed_env)
            if msr < satisfaction_rate:
                msr = satisfaction_rate
                max_cluster = cluster
                max_transformed_env = transformed_env
        working_cluster = copy.deepcopy(max_cluster)
        max_cluster_list = refactor_max_cluster(max_cluster)
        msr_opt, max_transformed_env_opt = 1.0, None
        for transform in max_cluster_list:
            temp_cluster = copy.deepcopy(max_cluster)
            temp_cluster[transform[0]][transform[1]].remove(transform[2])
            if len(temp_cluster[transform[0]][transform[1]]) == 0:
                del temp_cluster[transform[0]][transform[1]]
                if len(temp_cluster[transform[0]]) == 0:
                    del temp_cluster[transform[0]]
            transformed_env = generate_transformed_env(temp_cluster, "", save=False, try_load_existing=False)
            transformed_envs2.append(transformed_env)
            satisfaction_rate = calculate_possible_satisfaction_rate(transformed_env, anticipated_policy)
            satisfaction_rates2.append(satisfaction_rate)
            if msr_opt <= satisfaction_rate:
                msr_opt = satisfaction_rate
                working_cluster = dict_intersection(working_cluster, temp_cluster)
                max_transformed_env_opt = transformed_env
                best_cluster = working_cluster
        k += 1
    return max_transformed_env_opt, best_cluster



from src.rlpe.Environments.SingleTaxiEnv.single_taxi_wrapper import *
from src.rlpe.save_load_utils import *

ACT = "action"
IDX = "idx"
VAL = "val"


class SingleTaxiTransformSearchEnv(SingleTaxiSimpleEnv):
    def __init__(self):
        super().__init__()
        self.action_translation_dict = {}
        new_act = self.num_actions - 1
        preconditions = load_pkl_file(PRECONDITIONS_PATH)
        all_single_transformed_envs = load_pkl_file(ALL_SMALL_TAXI_TRANSFORMED_ENVS_PATH)
        for act, act_info in preconditions.not_allowed_features.items():
            for pre_idx, pre_val in act_info.items():
                for val in pre_val:
                    new_act += 1
                    transform_name = f"{act}_{pre_idx}_{val}"
                    if transform_name not in all_single_transformed_envs.keys():
                        continue
                    transformed_env = all_single_transformed_envs[transform_name]
                    self.action_translation_dict[new_act] = transform_name
                    for s, s_info in self.P.items():
                        self.P[s][new_act] = transformed_env.P[s][act]
        self.num_actions = new_act + 1
        discrete.DiscreteEnv.__init__(self, self.num_states, self.num_actions, self.P, self.initial_state_distribution)


class SingleTaxiTransformAnticipatedSearchEnv(SingleTaxiSimpleEnv):
    def __init__(self, anticipated_policy):
        super().__init__()
        self.action_translation_dict = {}
        new_act = self.num_actions - 1
        basic_actions = [_ for _ in range(self.num_actions)]
        preconditions = load_pkl_file(PRECONDITIONS_PATH)
        all_single_transformed_envs = load_pkl_file(ALL_SMALL_TAXI_TRANSFORMED_ENVS_PATH)
        decoded_anticipated_states, anticipated_actions = anticipated_policy.keys(), anticipated_policy.values()
        # flatten_anticipated_actions = [item for sublist in anticipated_actions for item in sublist]
        for state, state_info in self.P.items():
            new_act = self.num_actions - 1
            align_with_anticipated, anticipated_act, _ = is_state_align_with_anticipated_policy(self, state,
                                                                                                anticipated_policy)
            for not_allowed_act, not_allowed_act_info in preconditions.not_allowed_features.items():
                for pre_idx, pre_val in not_allowed_act_info.items():
                    for val in pre_val:
                        transform_name = f"{not_allowed_act}_{pre_idx}_{val}"
                        if transform_name not in all_single_transformed_envs.keys():
                            continue
                        new_act += 1
                        transformed_env = all_single_transformed_envs[transform_name]
                        self.action_translation_dict[new_act] = {ACT: not_allowed_act, IDX: pre_idx, VAL: val}
                        if align_with_anticipated:
                            temp_info = list(transformed_env.P[state][not_allowed_act][0])
                            if new_act == 24 and state == 8950:  # TODO - its a patch! to delete!
                                temp_info[3] = True
                                self.P[state][new_act] = [tuple(temp_info)]
                            else:
                                self.P[state][new_act] = transformed_env.P[state][not_allowed_act]
                        else:
                            temp_info = list(transformed_env.P[state][not_allowed_act][0])
                            temp_info[1] = state
                            temp_info[2] = temp_info[2] if temp_info[2] < 0 else -1.0
                            self.P[state][new_act] = [tuple(temp_info)]

            if align_with_anticipated:
                next_anticipated_state = self.P[state][anticipated_act][0][1]
                if state != next_anticipated_state:
                    relevant_actions = basic_actions
                else:
                    relevant_actions = [k for (k, v) in self.action_translation_dict.items() if
                                        v[ACT] == anticipated_act]
                    relevant_actions.append(anticipated_act)
                for other_act, other_act_info in self.P[state].items():
                    if other_act not in relevant_actions:
                        temp_info = list(other_act_info[0])
                        temp_info[1] = state
                        temp_info[2] = temp_info[2] if temp_info[2] < 0 else -1.0
                        self.P[state][other_act] = [tuple(temp_info)]
        self.num_actions = new_act + 1
        discrete.DiscreteEnv.__init__(self, self.num_states, self.num_actions, self.P, self.initial_state_distribution)


def create_transform_search_taxi_env(env, env_preconditions, anticipated_policy):
    default_data = [1.0, None, -1, False]
    translation_action_to_precondition_dict = dict()
    new_act_num = env.num_actions - 1
    for action in range(env.num_actions):
        action_preconditions = env_preconditions[action]
        for pre in action_preconditions:
            new_act_num += 1
            for s in env.P.keys():
                default_data[NEXT_STATE_IDX] = s
                env.P[s][new_act_num] = [tuple(default_data)]
                translation_action_to_precondition_dict[pre[PRECONDITION_NAME_IDX]] = new_act_num
    env.num_actions = new_act_num + 1
    discrete.DiscreteEnv.__init__(env, env.num_states, env.num_actions, env.P, env.initial_state_distribution)
    for anticipated_s, anticipated_actions in anticipated_policy.items():
        anticipated_states = env.get_states_from_partial_obs(anticipated_s)
        for anticipated_a in anticipated_actions:
            for anticipated_state in anticipated_states:
                tmp_info = list(env.P[anticipated_state][anticipated_a][0])
                next_state = tmp_info[NEXT_STATE_IDX]
                if next_state != anticipated_state:
                    for act in range(env.num_actions):
                        if act not in anticipated_actions:
                            tmp_info = list(env.P[anticipated_state][act][0])
                            tmp_info[NEXT_STATE_IDX] = anticipated_state
                            env.P[anticipated_state][act] = [tuple(tmp_info)]
                else:
                    relevant_preconditions = env_preconditions[anticipated_a]
                    for precondition in relevant_preconditions:
                        _, transformed_env = load_transform_by_name(precondition[PRECONDITION_NAME_IDX])
                        new_pre_action = translation_action_to_precondition_dict[precondition[PRECONDITION_NAME_IDX]]
                        env.P[anticipated_state][new_pre_action] = transformed_env.P[anticipated_state][anticipated_a]
    env.translation_action_to_precondition_dict = translation_action_to_precondition_dict
    return env


def reformat_preconditions(env_preconditions):
    new_preconditions = dict()
    for action, action_info in env_preconditions.not_allowed_features.items():
        for idx, values in action_info.items():
            for val in values:
                precondition_name = f"{action}_{idx}_{val}"
                if action in new_preconditions.keys():
                    new_preconditions[action].append((precondition_name, idx, val))
                else:
                    new_preconditions[action] = [(precondition_name, idx, val)]
    return new_preconditions

    # if __name__ == '__main__':
#     anticipated_policy = ANTICIPATED_POLICY
#     search_env = SingleTaxiTransformAnticipatedSearchEnv(anticipated_policy)
# file_name = "taxi_example_data/taxi_transformed_env/search_env_without_5.pkl"
# save_pkl_file(file_name, search_env)
#
# transform_name = "search_anticipated_env"
# save_pkl_file(transform_name, search_env)
# file = open(file_name, "rb")
# new_env = pickle.load(file)
# generate_agent(SINGLE_TAXI_EXAMPLE, KERAS_DQN, 100, new_env, transform_name)
# actions = [1, 1, 19, 7, 9, 9, 7, 24]
# all_reward = 0
# for act in actions:
#     search_env.render()
#     next_s, r, d, prob = search_env.step(act)
#     all_reward += r
#     print(f"state:{search_env.decode(next_s)}")
#     print(f"reward:{r} done:{d} prob:{prob}")
#     print(f"all_reward:{all_reward}")
# a = 7



import numpy as np
from queue import Queue

# produce a shortest path tree for the shortest ways to get to the location specified
# loc: tuple (row, col) of desired location  
# desc: description of environment map (env.desc)
from src.rlpe.Observer.abstract_expert import AbstractExpert


def shortest_path_tree(desc, loc):
    (loc_r, loc_c) = loc
    num_rows = len(desc) - 2
    num_cols = len(desc[0][1:-1:2])
    max_row = num_rows - 1
    max_col = num_cols - 1
    spt = [[[[], np.inf] for _ in range(num_cols)] for _ in range(num_rows)]

    spt[loc_r][loc_c][0].append(-1)  # after arriving at dest, no action should be taken
    spt[loc_r][loc_c][1] = 0

    q = Queue()
    q.put([loc, 0])
    while not q.empty():
        [(row, col), cur_dist] = q.get()  # current location we are exploring the neighbors of
        for action in range(4):
            if action == 0:
                (next_r, next_c) = (min(row + 1, max_row), col)
                if spt[next_r][next_c][1] >= cur_dist + 1:
                    if spt[next_r][next_c][1] > cur_dist + 1:
                        spt[next_r][next_c][1] = cur_dist + 1
                        q.put([(next_r, next_c), cur_dist + 1])
                    spt[next_r][next_c][0].append(1)  # add to optimal actions list

            elif action == 1:
                (next_r, next_c) = (max(row - 1, 0), col)
                if spt[next_r][next_c][1] >= cur_dist + 1:
                    if spt[next_r][next_c][1] > cur_dist + 1:
                        spt[next_r][next_c][1] = cur_dist + 1
                        q.put([(next_r, next_c), cur_dist + 1])
                    spt[next_r][next_c][0].append(0)  # add to optimal actions list

            elif action == 2 and desc[1 + row, 2 * col + 2] == b":":
                (next_r, next_c) = (row, min(col + 1, max_col))

                if spt[next_r][next_c][1] >= cur_dist + 1:
                    if spt[next_r][next_c][1] > cur_dist + 1:
                        spt[next_r][next_c][1] = cur_dist + 1
                        q.put([(next_r, next_c), cur_dist + 1])
                    spt[next_r][next_c][0].append(3)  # add to optimal actions list

            elif action == 3 and desc[1 + row, 2 * col] == b":":
                (next_r, next_c) = (row, max(col - 1, 0))

                if spt[next_r][next_c][1] >= cur_dist + 1:
                    if spt[next_r][next_c][1] > cur_dist + 1:
                        spt[next_r][next_c][1] = cur_dist + 1
                        q.put([(next_r, next_c), cur_dist + 1])
                    spt[next_r][next_c][0].append(2)  # add to optimal actions list

    return spt


# Hard-coded expert policy for Taxi-v2 domain
# Finds a policy set: a list of actions that would all count as part of the optimal policy

# Taxi Actions:
# There are 6 discrete deterministic actions:
# - 0: move south
# - 1: move north
# - 2: move east 
# - 3: move west 
# - 4: pickup passenger
# - 5: dropoff passenger


class Taxi_Expert(AbstractExpert):
    def __init__(self, env):
        self.env = env
        self.desc = env.desc  # array holding map layout
        self.locs = env.passengers_locations  # locations of passenger pickup or dropoff X's on map
        self.num_rows = len(self.desc) - 2
        self.num_cols = len(self.desc[0][1:-1:2])
        self.max_row = self.num_rows - 1
        self.max_col = self.num_cols - 1
        # self.shortest_path_trees = []
        # for loc in self.locs:
        #     self.shortest_path_trees.append(shortest_path_tree(self.desc, loc))
        self.shortest_path_trees = {}
        for loc in env.passengers_locations:
            self.shortest_path_trees[tuple(loc)] = shortest_path_tree(self.desc, loc)
        # self.ignored_taxi_locations = [(0, 1), (1, 1), (0, 2), (1, 2), (0, 3), (1, 3), (0, 4), (1, 4)]
        self.ignored_taxi_locations = []

    # get expert action, given a tuple of the form:
    # (taxi_loc_x, taxi_loc_y, fuel_level, pass_start_x, pass_start_y, pass_dest_x, pass_dest_y, pass_status)
    # unnecessary information in the tuple can be left as None (e.g. fuel level is not used in this function and can be None)
    def get_expert_policy_set(self, state_tuple):
        (taxi_loc_x, taxi_loc_y, fuel_level, pass_start_x, pass_start_y, pass_dest_x, pass_dest_y,
         pass_status) = state_tuple
        taxi_loc = (taxi_loc_x, taxi_loc_y)
        pass_start = (pass_start_x, pass_start_y)
        pass_dest = (pass_dest_x, pass_dest_y)
        if pass_status == 3 and taxi_loc == pass_dest:
            return [5]  # dropoff
        elif pass_status == 3 and taxi_loc != pass_dest:
            return self.shortest_path_trees[pass_dest][taxi_loc_x][taxi_loc_y][0]
        elif pass_status == 2 and taxi_loc == pass_start:
            return [4]  # pickup
        elif pass_status == 2 and taxi_loc != pass_start:
            return self.shortest_path_trees[pass_start][taxi_loc_x][taxi_loc_y][0]
        elif pass_status == 1:
            print("Passenger has already arrived")
            return None
        else:
            print("Not a valid state")

    def full_expert_policy_dict(self):
        fuel_level = None
        expert_policy_dict = {}
        for taxi_loc_x in range(self.num_rows):
            for taxi_loc_y in range(self.num_cols):
                if (taxi_loc_x, taxi_loc_y) in self.ignored_taxi_locations:
                    continue
                for pass_start in self.env.passengers_start_fixed_locations:
                    for pass_dest in self.env.passengers_fixed_destinations:
                        for pass_status in [2, 3]:
                            state_tuple = (
                                taxi_loc_x, taxi_loc_y, fuel_level, pass_start[0], pass_start[1], pass_dest[0],
                                pass_dest[1], pass_status)
                            expert_policy_dict[state_tuple] = self.get_expert_policy_set(state_tuple)
        return expert_policy_dict





# ================= Train variables ================ #
ITER_NUM = 1000000
# =================== Transforms =================== #
FUELS_TRANSFORM = "fuels_transform"
REWARD_TRANSFORM = "reward_transform"
NO_WALLS_TRANSFORM = "no_walls_transform"
WITHOUT_TRANSFORM = "Without transform"
# ================== Environments ================== #
TAXI = "taxi"
SPEAKER_LISTENER = "simple_speaker_listener"
TAXI_STATES_NUM = 5000
SEARCH_TRANSFORM_TAXI_ENV = "big_taxi_search_transform_env"
APPLE_PICKING = "apple_picking_env"
# ================= Env variables ================== #

TRAIN = "train"
EVALUATE = "evaluate"
EPISODE_REWARD_MEAN = "episode_reward_mean"
EPISODE_REWARD_MAX = "episode_reward_max"
EPISODE_REWARD_MIN = "episode_reward_min"
EPISODE_STEP_NUM_MEAN = "episode_len_mean"
EPISODE_VARIANCE = "episode_variance"
TOTAL_EPISODE_REWARD = "total_episode_reward"
TRAINING_RESULTS = "training_results"
EVALUATION_RESULTS = "evaluation_results"
SUCCESS_RATE = "success_rate"
GOT_AN_EXPLANATION = "explanation"
ORIGINAL_ENV = "original_env"
FUEL_TRANSFORMED_ENV = "fuel"

SMALL_TAXI_ANTICIPATED_POLICY = {(2, 0, 0, 3, None): [1],
                                 (1, 0, 0, 3, None): [1],
                                 (0, 0, 0, 3, None): [4],
                                 (0, 0, 4, 3, None): [0],
                                 (1, 0, 4, 3, None): [2],
                                 (1, 1, 4, 3, None): [2],
                                 (1, 2, 4, 3, None): [0],
                                 (2, 2, 4, 3, None): [5]}

BIG_TAXI_ANTICIPATED_POLICY = {(4, 0, 0, 3, None): [1],
                               (3, 0, 0, 3, None): [1],
                               (2, 0, 0, 3, None): [1],
                               (1, 0, 0, 3, None): [1],
                               (0, 0, 0, 3, None): [4],
                               (0, 0, 4, 3, None): [0, 2],
                               (0, 1, 4, 3, None): [0],
                               (1, 0, 4, 3, None): [0, 2],
                               (1, 1, 4, 3, None): [0],
                               (2, 0, 4, 3, None): [2],
                               (2, 1, 4, 3, None): [2],
                               (2, 2, 4, 3, None): [2],
                               (2, 3, 4, 3, None): [0, 2],
                               (2, 4, 4, 3, None): [0],
                               (3, 3, 4, 3, None): [0, 2],
                               (3, 4, 4, 3, None): [0],
                               (4, 3, 4, 3, None): [2],
                               (4, 4, 4, 3, None): [5]
                               }
ANTICIPATED_POLICY = BIG_TAXI_ANTICIPATED_POLICY

PRECONDITION_NAME_IDX = 0
NEXT_STATE_IDX = 1
# ---------------------- PATHS ----------------------- #
CUR_ENV_NAME = "big_taxi"

DATA_FOLDER = "Data/"
TRAINED_AGENTS_DIR_PATH = DATA_FOLDER + "TrainedAgentsDQN/"
TRAINED_AGENT_SAVE_PATH = TRAINED_AGENTS_DIR_PATH + f"trained_models_on_{CUR_ENV_NAME}_env/"
TRAINED_AGENT_SPECIFIC_PATH = "single_transform_envs/"
TRAINED_AGENT_ON_SPECIFIC_TRANSFORM_NUM_PATH = TRAINED_AGENT_SAVE_PATH + TRAINED_AGENT_SPECIFIC_PATH
TRAINED_AGENT_RESULTS_PATH = TRAINED_AGENTS_DIR_PATH + f"results/trained_models_on_{CUR_ENV_NAME}_env_results/"
TRAINED_AGENT_RESULT_FILE_PATH = TRAINED_AGENT_RESULTS_PATH + f"results_{CUR_ENV_NAME}_search_transform.pkl"

TAXI_TRANSFORM_DATA_PATH = DATA_FOLDER + f"{CUR_ENV_NAME}_env_data/"
PRECONDITIONS_FILE_NAME = f"{CUR_ENV_NAME}_env_preconditions.pkl"
PRECONDITIONS_PATH = TAXI_TRANSFORM_DATA_PATH + PRECONDITIONS_FILE_NAME
PRECONDITION_GRAPH_PATH = TAXI_TRANSFORM_DATA_PATH + f"precondition_graph_{CUR_ENV_NAME}_env.pkl"
ALL_SMALL_TAXI_TRANSFORMED_ENVS_PATH = TAXI_TRANSFORM_DATA_PATH + "all_single_transformed_envs.pkl"
TRANSFORMS_PATH = TAXI_TRANSFORM_DATA_PATH + f"{CUR_ENV_NAME}_transformed_envs/"

TRANSFORM_SEARCH_TAXI_ENV_FILENAME = f"{CUR_ENV_NAME}_search_transform_env.pkl"
TRANSFORM_SEARCH_TAXI_ENV_PATH = TAXI_TRANSFORM_DATA_PATH + TRANSFORM_SEARCH_TAXI_ENV_FILENAME

PATH_TO_DATA_FOR_COLAB = DATA_FOLDER + "collab_example_data/"
COLAB_TRANSFORMS = PATH_TO_DATA_FOR_COLAB + f"{CUR_ENV_NAME}_transformed_env_colab/"


from src.rlpe.Agents.RL_agents.q_learning_agents import *
from src.rlpe.Agents.RL_agents.sarsa_agents import *

from src.rlpe.Agents.value_iteration_agent import ValueIterationAgent

FORMAT_STRING = "\r{:3d} mean reward: {:6.2f}, variance: {:6.2f}, number of steps: {}"
from src.rlpe.constants import *


def print_info(it, episode_result):
    print(FORMAT_STRING.format(it + 1, episode_result[EPISODE_REWARD_MEAN], episode_result[EPISODE_VARIANCE],
                               episode_result[EPISODE_STEP_NUM_MEAN]))


def get_rl_agent(agent_name, env, callbacks=None,  env_name=None, env_to_agent=None):  # TODO - fix the environment
    if is_rllib_agent(agent_name):
        agent = get_rllib_agent(agent_name, env_name, env, env_to_agent)
    elif agent_name == HANDS_ON_DQN:
        agent = HandsOnDQNAgent(env=env)
    elif agent_name == KERAS_DQN:
        agent = DQNKeras(env=env, callbacks=callbacks)
    elif agent_name == Q_LEARNING:
        agent = QLearningAgent(env=env)
    elif agent_name == KERAS_SARSA:
        agent = KerasSarsaAgent(env=env)
    elif agent_name == KERAS_CEM:
        agent = KerasCEMAgent(env=env)
    elif agent_name == VALUE_ITERATION:
        agent = ValueIterationAgent(env=env)
    else:
        raise Exception("Not valid agent name")
    return agent


def run(agent, num_of_episodes, method=TRAIN, print_process=True, visualize=True):
    result = {EPISODE_REWARD_MEAN: np.zeros(num_of_episodes), EPISODE_REWARD_MAX: np.zeros(num_of_episodes),
              EPISODE_REWARD_MIN: np.zeros(num_of_episodes), EPISODE_STEP_NUM_MEAN: np.zeros(num_of_episodes),
              EPISODE_VARIANCE: np.zeros(num_of_episodes)}
    print_rate = 100
    if print_process:
        message = "Training:" if method == TRAIN else "\nEvaluating:"
        print(message)
    if not isinstance(agent, DQNKeras): # and not isinstance(agent, KerasSarsaAgent): TODO - to add this!
        for it in range(num_of_episodes):
            episode_result = agent.run()
            if print_process and (it + 1) % print_rate == 0:
                print("\rEpisode {}/{}.".format(it + 1, num_of_episodes), end="")
                sys.stdout.flush()

            result[EPISODE_REWARD_MEAN][it] = sum(episode_result[EPISODE_REWARD_MEAN]) / len(episode_result[EPISODE_REWARD_MEAN])
            result[EPISODE_REWARD_MAX][it] = (episode_result[EPISODE_REWARD_MAX])
            result[EPISODE_REWARD_MIN][it] = (episode_result[EPISODE_REWARD_MIN])
            result[EPISODE_STEP_NUM_MEAN][it] = sum(episode_result[EPISODE_STEP_NUM_MEAN]) / len(episode_result[EPISODE_STEP_NUM_MEAN])
            result[EPISODE_VARIANCE][it] = (episode_result[EPISODE_VARIANCE])
        if method == EVALUATE:
            agent.evaluate(visualize)
    else:
        # result = agent.run()
        if method == EVALUATE:
            agent.evaluate(visualize)
        else:
            result = agent.run()
    return result


def run_episode(env, agent, method=TRAIN):
    state = env.reset()

    # == Initialize variables == #
    result = {EPISODE_REWARD_MEAN: 0.0, EPISODE_REWARD_MIN: np.inf, EPISODE_REWARD_MAX: -np.inf,
              EPISODE_STEP_NUM_MEAN: 0, EPISODE_VARIANCE: 0.0, TOTAL_EPISODE_REWARD: 0.0}
    total_reward = 0.0
    episode_len = 0
    for timestep in range(agent.timesteps_per_episode):
        if method == EVALUATE:
            env.render()
            pass

        action = agent.compute_action(state)  # Run Action

        next_state, reward, done, info = agent.env.step(action)  # Take action
        if method == EVALUATE:
            print(f"reward: {reward}, done: {done}, info: {info}")
        episode_len += 1

        total_reward += reward
        result[EPISODE_REWARD_MAX] = reward if reward > result[EPISODE_REWARD_MAX] else result[EPISODE_REWARD_MAX]
        result[EPISODE_REWARD_MIN] = reward if reward < result[EPISODE_REWARD_MIN] else result[EPISODE_REWARD_MIN]

        terminated = done
        if terminated:
            agent.stop_episode()
            break

        state = agent.episode_callback(state, action, reward, next_state, terminated)

    result[TOTAL_EPISODE_REWARD] = total_reward
    result[EPISODE_REWARD_MEAN] = total_reward / agent.timesteps_per_episode
    result[EPISODE_STEP_NUM_MEAN] = episode_len
    result[EPISODE_VARIANCE] = result[EPISODE_REWARD_MAX] - result[EPISODE_REWARD_MIN]
    return result


def create_agent(env, agent_name, callbacks=None, env_name=None):
    agent = get_rl_agent(agent_name, env, callbacks)
    return agent


# get the action performed by the agents in each observation
def get_policy_action(agent_rep, obs, reshape=False):
    # [taxi location], [current_fuel], [passengers_start_locations], [destinations], [passengers_status]
    if reshape:
        obs = np.reshape(obs, (1, len(obs)))
    action = agent_rep.compute_action(obs)
    return action


def get_policy_action_partial_obs(agent_rep, partial_obs, reshape=False):
    # [taxi location], [current_fuel], [passengers_start_locations], [destinations], [passengers_status]
    if reshape:
        partial_obs = np.reshape(partial_obs, (1, len(partial_obs)))
    action = agent_rep.compute_action(partial_obs)
    return action



from src.rlpe.Environments.SingleTaxiEnv.single_taxi_wrapper import *
from src.rlpe.Transforms.env_transforms import *

DETERMINISTIC = True


class SingleTaxiTransformedEnv(SingleTaxiSimpleEnv):
    def __init__(self, transforms):
        super().__init__(DETERMINISTIC)
        self.state_visibility_transforms = transforms[STATE_VISIBILITY_TRANSFORM]
        self.all_outcome_determinization = transforms[ALL_OUTCOME_DETERMINIZATION]
        self.most_likely_outcome = transforms[MOST_LIKELY_OUTCOME]
        self.relax_preconditions = transforms[PRECONDITION_RELAXATION]

        if self.all_outcome_determinization:
            get_all_outcome_determinization_matrix(self.P)
        if self.most_likely_outcome:
            self.P = get_most_likely_outcome(self.P)
        if self.relax_preconditions:
            self.preconditions = preconditions_relaxation(self.relax_preconditions, self)
            self.preconditions_num = sum([len(self.preconditions.not_allowed_features[k]) for k in
                                          self.preconditions.not_allowed_features.keys()])

    def step(self, a):
        s, r, d, p = super(SingleTaxiTransformedEnv, self).step(a)
        next_state = self.decode(s)
        env_default_values = self.state_visibility_transforms[1]
        for i in self.state_visibility_transforms[0]:
            next_state[i] = env_default_values[i]

        transformed_next_state = self.encode(*next_state)
        return int(transformed_next_state), r, d, p


def get_all_outcome_determinization_matrix(p):
    # this is the function that converts the matrix!
    mapped_actions = {}
    ABNORMAL_STATES = [15]
    # states like the goal state, where the results are different, should be ignored in the action mapping
    # ABNORMAL_STATES should also include hole states but for now I have not included them because in this
    # particular case it doesn't matter

    for (s, s_outcomes) in p.items():
        new_outcomes = {}
        i = 0
        for (a, a_outcomes) in s_outcomes.items():
            for probs in a_outcomes:
                temp = (1.0,) + probs[1:]
                new_outcomes[i] = temp
                if s not in ABNORMAL_STATES:
                    mapped_actions[i] = a
                i += 1
        p[s] = new_outcomes


def get_single_taxi_transform_name(transforms):
    taxi_x_transform, taxi_y_transform = transforms[0], transforms[1]
    pass_loc_transform, pass_dest_transform = transforms[2], transforms[3]
    fuel_transform = transforms[4]
    all_outcome_determinization = transforms[5]
    most_likely_outcome = transforms[6]

    name = ""
    name += TAXI_LOC_X if taxi_x_transform else ""
    name += TAXI_LOC_Y if taxi_y_transform else ""
    name += PASS_LOC if pass_loc_transform else ""
    name += PASS_DEST if pass_dest_transform else ""
    name += FUEL if fuel_transform else ""
    name += ALL_OUTCOME_DETERMINIZATION if all_outcome_determinization else ""
    name += MOST_LIKELY_OUTCOME if most_likely_outcome else ""
    return name


def generate_triple_of_transforms(env_pre):
    dir_name = "triple_transform_envs"
    make_dir(TRAINED_AGENT_SAVE_PATH + dir_name)

    same_precondition = False
    for act1, pre1 in env_pre.not_allowed_features.items():
        for act2, pre2 in env_pre.not_allowed_features.items():
            for act3, pre3 in env_pre.not_allowed_features.items():
                for pre_idx1 in pre1.keys():
                    for pre_idx2 in pre2.keys():
                        for pre_idx3 in pre3.keys():
                            for pre_val1 in pre1[pre_idx1]:
                                for pre_val2 in pre2[pre_idx2]:
                                    for pre_val3 in pre3[pre_idx3]:
                                        if (act1 == act2 and (np.array(pre_idx1 == pre_idx2)).all() and np.array(
                                                (pre_val1 == pre_val2)).all()) or (act1 > act2) or ((act2 == act3 and (
                                                np.array(pre_idx2 == pre_idx3)).all() and np.array(
                                            (pre_val2 == pre_val3)).all()) or (act2 > act3)) or ((act1 == act3 and (
                                                np.array(pre_idx1 == pre_idx3)).all() and np.array(
                                            (pre_val1 == pre_val3)).all()) or (act1 > act3)):
                                            same_precondition = True
                                        if not same_precondition:

                                            env_file_name = f"{TRAINED_AGENT_SAVE_PATH}{dir_name}/{act1}_{pre_idx1}_{pre_val1}_{act2}_{pre_idx2}_{pre_val2}_{act3}_{pre_idx3}_{pre_val3}"
                                            if not os.path.exists(env_file_name + ".pkl"):
                                                print(f"act1: {act1} , pre_idx1: {pre_idx1} , pre_val1: {pre_val1}")
                                                print(f"act2: {act2} , pre_idx2: {pre_idx2} , pre_val2: {pre_val2}")
                                                print(f"act3: {act3} , pre_idx3: {pre_idx3} , pre_val3: {pre_val3}")
                                                print("\n")
                                                precondition = {act1: {pre_idx1: pre_val1},
                                                                act2: {pre_idx2: pre_val2},
                                                                act3: {pre_idx3: pre_val3}}
                                                generate_transformed_env(precondition, env_file_name, save=True)
                                            else:
                                                print(f"file: {env_file_name} already exists")
                                        same_precondition = False


def generate_double_of_transforms(env_pre):
    dir_name = "double_transform_envs"
    make_dir(TRAINED_AGENT_SAVE_PATH + dir_name)

    same_precondition = False
    for act1, pre1 in env_pre.not_allowed_features.items():
        for act2, pre2 in env_pre.not_allowed_features.items():
            for pre_idx1 in pre1.keys():
                for pre_idx2 in pre2.keys():
                    for pre_val1 in pre1[pre_idx1]:
                        for pre_val2 in pre2[pre_idx2]:
                            if (act1 == act2 and (np.array(pre_idx1 == pre_idx2)).all() and np.array(
                                    (pre_val1 == pre_val2)).all()) or (act1 > act2):
                                same_precondition = True
                            if not same_precondition:
                                print(f"act1: {act1} , pre_idx1: {pre_idx1} , pre_val1: {pre_val1}")
                                print(f"act2: {act2} , pre_idx2: {pre_idx2} , pre_val2: {pre_val2}")

                                env_file_name = f"{TRAINED_AGENT_SAVE_PATH}{dir_name}/{act1}_{pre_idx1}_{pre_val1}_{act2}_{pre_idx2}_{pre_val2}"
                                precondition = {act1: {pre_idx1: pre_val1},
                                                act2: {pre_idx2: pre_val2}}
                                generate_transformed_env(precondition, env_file_name, save=True)
                            same_precondition = False


def generate_single_transforms(env_preconditions):
    dir_name = "single_transform_envs"
    make_dir(TRAINED_AGENT_SAVE_PATH + dir_name)

    for act, preconditions in env_preconditions.not_allowed_features.items():
        for pre_idx in preconditions.keys():
            for pre_val in preconditions[pre_idx]:
                print(
                    f"calculating for act1: {act} , pre_idx1: {pre_idx} , pre_val1: {pre_val}")
                env_file_name = f"{TRAINED_AGENT_SAVE_PATH}{dir_name}/{act}_{pre_idx}_{pre_val}"
                precondition = {act: {pre_idx: pre_val}}
                generate_transformed_env(precondition, env_file_name, save=True)


def generate_transforms_to_depth(preconditions, anticipated_policy, max_depth=None):
    if not max_depth:
        max_depth = len(anticipated_policy)
    basic_preconditions_list = []
    for act, action_info in preconditions.items():
        for idx, values in action_info.items():
            for val in values:
                pre_string = f"{act}_{idx}_{val}"
                precondition_actions = get_precondition_actions_from_string(pre_string)
                influence = is_transform_actions_influence_anticipated_policy(list(anticipated_policy.values()),
                                                                              precondition_actions)
                if influence:
                    basic_preconditions_list.append(pre_string)
                    precondition = {act: {idx: val}}
                    env_file_name_path = f"{TRANSFORMS_PATH}" + pre_string
                    generate_transformed_env(precondition, env_file_name_path, try_load_existing=False, save=True)
    # basic_preconditions_list = load_pkl_file("basic_preconditions_list.pkl")

    all_preconditions_list = {1: tuple(tuple([pre]) for pre in basic_preconditions_list)}
    for i in range(2, max_depth):
        print(f"depth: {i}")
        all_preconditions_list[i] = tuple(itertools.combinations(basic_preconditions_list, i))
        for pre in all_preconditions_list[i]:
            precondition = dict()
            env_file_name = ""
            for p in pre:
                act = get_precondition_actions_from_string(p)[0]
                idx = get_precondition_idx_from_string(p)
                val = get_precondition_val_from_string(p)
                env_file_name += f"{act}_{idx}_{val}_"
                if act in precondition.keys():
                    if idx in precondition[act].keys():
                        precondition[act][idx].append(val)
                    else:
                        precondition[act][idx] = [val]
                else:
                    precondition[act] = {idx: [val]}
            env_file_name = env_file_name[:-1]
            env_file_name_path = f"{TRANSFORMS_PATH}" + env_file_name
            generate_transformed_env(precondition, env_file_name_path, save=True)


def generate_transformed_env(precondition, env_file_name, save=True, try_load_existing=True, env_default_values=None):
    print(f"Generating precondition env: {precondition}")
    new_env = None
    if try_load_existing:
        file_name = os.path.basename(env_file_name)
        transform_name, new_env = load_transform_by_name(file_name)
    if new_env is None:
        cur_transforms = {STATE_VISIBILITY_TRANSFORM: ([], env_default_values),
                          ALL_OUTCOME_DETERMINIZATION: False,
                          MOST_LIKELY_OUTCOME: False,
                          PRECONDITION_RELAXATION: precondition}
        new_env = SingleTaxiTransformedEnv(cur_transforms)
        if save:
            save_pkl_file(env_file_name + ".pkl", new_env)
    return new_env

# if __name__ == '__main__':
#     #     env_default_values = [0, 0, 0, 1, MAX_FUEL - 1]
#     a_file = open("taxi_example_data/small_taxi_env_preconditions.pkl", "rb")
#     cur_env_preconditions = pickle.load(a_file)
#
#     act1, act2, act3, act4, act5 = 0, 1, 2, 4, 5
#     pre_idx = (4,)
#     pre_val = [0]
#     dir_name = "taxi_example_data/taxi_transformed_env/"
#     env_file_name = f"{act1}_{pre_idx}_{pre_val}_{act2}_{pre_idx}_{pre_val}_{act3}_{pre_idx}_{pre_val}_{act4}_{pre_idx}_{pre_val}_{act5}_{pre_idx}_{pre_val}.pkl"
#     precondition = {act1: {pre_idx: pre_val},
#                     act2: {pre_idx: pre_val},
#                     act3: {pre_idx: pre_val},
#                     act4: {pre_idx: pre_val},
#                     act5: {pre_idx: pre_val},
#                     }
#     env = generate_transformed_env(precondition, env_file_name=dir_name + env_file_name, save=True)
#
#     print("DONE!")



from src.rlpe.Environments.LunarLander.lunar_lander import *
from src.rlpe.Environments.abstract_wrapper_env import AbstractWrapperEnv
import json
from itertools import product


def state_extractor(s):
    state = (min(2, max(-2, int((s[0]) / 0.05))),
             min(2, max(-1, int((s[1]) / 0.1))),
             min(2, max(-2, int((s[2]) / 0.1))),
             min(2, max(-2, int((s[3]) / 0.1))),
             min(2, max(-2, int((s[4]) / 0.1))),
             min(2, max(-2, int((s[5]) / 0.1))),
             int(s[6]),
             int(s[7]))
    return state


class LunarLenderWrapper(LunarLander, AbstractWrapperEnv):
    def __init__(self):
        super().__init__()
        with open('results/sarsa_data/sarsa_Q_5X4Ys.json') as json_file:
            Q = json.load(json_file)
            self.num_states = len(Q)
        self.transform_num = 6

    def reset(self):
        obs = super(LunarLenderWrapper, self).reset()
        encoded_state = self.encode(obs) if not isinstance(obs, int) else obs
        decoded_state = self.decode(encoded_state)
        return encoded_state

    def step(self, action):
        next_state, reward, done, info = super(LunarLenderWrapper, self).step(action)
        encoded_next_state = self.encode(next_state)
        decoded_state = self.decode(encoded_next_state)
        return encoded_next_state, reward, done, info

    def get_states_from_partial_obs(self, partial_obs):
        partial_obs = list(partial_obs)
        x_coordinate = [partial_obs[0]] if (partial_obs[0] is not None) else list(range(-2, 3))
        y_coordinate = [partial_obs[1]] if (partial_obs[1] is not None) else list(range(-2, 3))
        x_horizontal_velocity = [partial_obs[2]] if (partial_obs[2] is not None) else list(range(-2, 3))
        y_vertical_velocity = [partial_obs[3]] if (partial_obs[3] is not None) else list(range(-2, 3))
        orientation_in_space = [partial_obs[4]] if (partial_obs[4] is not None) else list(range(-2, 3))
        angular_velocity = [partial_obs[5]] if (partial_obs[5] is not None) else list(range(-2, 3))
        left_leg = [partial_obs[6]] if (partial_obs[6] is not None) else list(range(2))
        right_leg = [partial_obs[7]] if (partial_obs[7] is not None) else list(range(2))
        states = list(
            product(x_coordinate, y_coordinate, x_horizontal_velocity, y_vertical_velocity, orientation_in_space,
                    angular_velocity, left_leg, right_leg, repeat=1))
        states = [self.encode(state) for state in states]
        return states

    def encode(self, state):
        state = state_extractor(state)
        s = [(state[i] + 2) for i in range(len(state)) if i < 6] + [state[6]] + [state[7]]
        x_coordinate, y_coordinate = s[0], s[1]
        x_horizontal_velocity, y_vertical_velocity = s[2], s[3]
        orientation_in_space = s[4]
        angular_velocity = s[5]
        left_leg, right_leg = s[6], s[7]

        i = x_coordinate

        i *= 5
        i += y_coordinate

        i *= 5
        i += x_horizontal_velocity

        i *= 5
        i += y_vertical_velocity

        i *= 5
        i += orientation_in_space

        i *= 5
        i += angular_velocity

        i *= 2
        i += left_leg

        i *= 2
        i += right_leg
        return i

    def decode(self, i):
        j = i
        out = []

        right_leg = i % 2
        out.append(right_leg)
        i = i // 2

        left_leg = i % 2
        out.append(left_leg)
        i = i // 2

        angular_velocity = i % 5
        out.append(angular_velocity)
        i = i // 5

        orientation_in_space = i % 5
        out.append(orientation_in_space)
        i = i // 5

        y_vertical_velocity = i % 5
        out.append(y_vertical_velocity)
        i = i // 5

        x_horizontal_velocity = i % 5
        out.append(x_horizontal_velocity)
        i = i // 5

        y_coordinate = i % 5
        out.append(y_coordinate)
        i = i // 5

        x_coordinate = i % 5
        out.append(x_coordinate)

        assert 0 <= i < 5
        state = list(reversed(out))
        state = [(state[i] - 2) for i in range(len(state)) if i < 6] + [state[6]] + [state[7]]
        return state



from abc import ABC, abstractmethod


class AbstractAgent(ABC):
    def __init__(self, env, timesteps_per_episode=10001):
        self.env = env
        self.timesteps_per_episode = timesteps_per_episode

    @abstractmethod
    def run(self) -> {str: float}:
        """
        The agent's training method.
        Returns: a dictionary - {"episode_reward_mean": __, "episode_reward_min": __, "episode_reward_max": __,
        "episode_len_mean": __}
        """
        pass

    @abstractmethod
    def compute_action(self, state) -> int:
        """
        Computes the best action from a given state.
        Returns: a int that represents the best action.
        """
        pass

    @abstractmethod
    def stop_episode(self):
        pass

    @abstractmethod
    def episode_callback(self, state, action, reward, next_state, terminated):
        pass

    @abstractmethod
    def evaluate(self):
        pass



SOUTH, NORTH, EAST, WEST, PICKUP = 0, 1, 2, 3, 4
ACTIONS = [SOUTH, NORTH, EAST, WEST, PICKUP]

STEP_REWARD, APPLE_PICKUP_REWARD, BAD_APPLE_PICKUP_REWARD, STEP_INTO_THORN_REWARD = "step", "good_pickup", "bad_pickup", "step_into_thorn"
REWARD_DICT = {STEP_REWARD: -1, APPLE_PICKUP_REWARD: 20, BAD_APPLE_PICKUP_REWARD: -10, STEP_INTO_THORN_REWARD: -5}

ABOVE = -1
UNDER = 1
COLLECTED = 1
NOT_COLLECTED = 0

DETERMINISTIC_PROB = 1.0
STOCHASTIC_PROB = 0.60
STOCHASTIC_PROB_THORNY = 1 - STOCHASTIC_PROB



from src.rlpe.Environments.MultiTaxiEnv.multitaxienv.config import TAXI_ENVIRONMENT_REWARDS
from src.rlpe.Environments.taxi_environment_wrapper import TaxiSimpleExampleEnv
from src.rlpe.constants import *
from src.rlpe.Transforms.transform_constants import *

SMALL_MAP = [
    "+-----+",
    "|X: :X|",
    "| : : |",
    "|X:F:X|",
    "+-----+",
]

# NEW_MAP = [
#     "+---------+",
#     "|X: :F: :X|",
#     "| : : : : |",
#     "| : : : : |",
#     "| : : : : |",
#     "|X: :G:X: |",
#     "+---------+",
# ]
TAXI_NAME = "taxi_1"
my_reward = dict(
    step=-1,
    no_fuel=-20,
    bad_pickup=-15,
    bad_dropoff=-15,
    bad_refuel=-10,
    bad_fuel=-50,
    pickup=50,
    standby_engine_off=-1,
    turn_engine_on=-10e6,
    turn_engine_off=-10e6,
    standby_engine_on=-1,
    intermediate_dropoff=-15,
    final_dropoff=100,
    hit_wall=-2,
    collision=-35,
    collided=-20,
    unrelated_action=-15
)


# fuel_transform=False, status_transform=False, reward_transform=False

class TaxiTransformedEnv(TaxiSimpleExampleEnv):
    def __init__(self, transforms):
        super().__init__(max_fuel=[3] if transforms[2] else [100])
        taxi_loc_x_transform, taxi_loc_y_transform = transforms[0], transforms[1]
        fuel_transform = transforms[2]
        pass_loc_x_transform, pass_loc_y_transform = transforms[3], transforms[4]
        pass_dest_x_transform, pass_dest_y_transform = transforms[5], transforms[6]
        pass_status_transform = transforms[7]
        no_walls_transform = transforms[8]
        reward_transform = transforms[9]

        self.taxi_loc_x_transform = taxi_loc_x_transform
        self.taxi_loc_y_transform = taxi_loc_y_transform
        self.fuel_transform = fuel_transform
        self.pass_loc_x_transform = pass_loc_x_transform
        self.pass_loc_y_transform = pass_loc_y_transform
        self.pass_dest_x_transform = pass_dest_x_transform
        self.pass_dest_y_transform = pass_dest_y_transform
        self.pass_status_transform = pass_status_transform
        self.no_walls_transform = no_walls_transform
        self.reward_transform = reward_transform
        self.reward_dict = my_reward if self.reward_transform else TAXI_ENVIRONMENT_REWARDS

    def _update_movement_wrt_fuel(self, taxi: int, taxis_locations: list, wanted_row: int, wanted_col: int,
                                  reward: int, fuel: int) -> (int, int, list):
        """
        Given that a taxi would like to move - check the fuel accordingly and update reward and location.
        Args:
            taxi: index of the taxi
            taxis_locations: list of current locations (prior to movement)
            wanted_row: row after movement
            wanted_col: col after movement
            reward: current reward
            fuel: current fuel

        Returns: updated_reward, updated fuel, updared_taxis_locations

        """
        reward = reward
        fuel = fuel
        taxis_locations = taxis_locations
        if fuel == 0:
            reward = ('no_fuel')
        else:
            if not self.fuel_transform:
                fuel = max(0, fuel - 1)

            taxis_locations[taxi] = [wanted_row, wanted_col]

        return reward, fuel, taxis_locations

    def set_reward_dict(self, new_rewards):
        self.reward_dict = new_rewards

    def step(self, action):
        """
        Executing a list of actions (action for each taxi) at the domain current state.
        Supports not-joined actions, just pass 1 element instead of list.
        Using manual reward.

        Args:
            action_dict: {taxi_name: action} - action of specific taxis to take on the step

        Returns: - dict{taxi_id: observation}, dict{taxi_id: reward}, dict{taxi_id: done}, _
        """
        action_dict = {TAXI_NAME: action}
        rewards = {}
        self.counter += 1
        if self.counter >= 90:
            self.window_size = 3

        # Main of the function, for each taxi-i act on action[i]
        for taxi_name, action_list in action_dict.items():
            # meta operations on the type of the action
            action_list = self._get_action_list(action_list)

            for action in action_list:
                taxi = self.taxis_names.index(taxi_name)
                reward = self.partial_closest_path_reward('step')  # Default reward
                moved = False  # Indicator variable for later use

                # taxi locations: [i, j]
                # fuels: int
                # passengers_start_locations and destinations: [[i, j] ... [i, j]]
                # passengers_status: [[1, 2, taxi_index+2] ... [1, 2, taxi_index+2]], 1 - delivered
                taxis_locations, fuels, passengers_start_locations, destinations, passengers_status = self.state

                if all(list(self.dones.values())):
                    continue

                # If taxi is collided, it can't perform a step
                if self.collided[taxi] == 1:
                    rewards[taxi_name] = self.partial_closest_path_reward('collided')
                    self.dones[taxi_name] = True
                    continue

                # If the taxi is out of fuel, it can't perform a step
                if fuels[taxi] == 0 and not self.at_valid_fuel_station(taxi, taxis_locations):
                    rewards[taxi_name] = self.partial_closest_path_reward('bad_fuel')
                    self.dones[taxi_name] = True
                    continue

                taxi_location = taxis_locations[taxi]
                row, col = taxi_location

                fuel = fuels[taxi]
                is_taxi_engine_on = self.engine_status_list[taxi]
                _, index_action_dictionary = self.get_available_actions_dictionary()

                if not is_taxi_engine_on:  # Engine is off
                    # update reward according to standby/ turn-on/ unrelated + turn engine on if requsted
                    reward = self._engine_is_off_actions(index_action_dictionary[action], taxi)

                else:  # Engine is on
                    # Binding
                    if index_action_dictionary[action] == 'bind':
                        self.bounded = False
                        reward = self.partial_closest_path_reward('bind')

                    # Movement
                    if index_action_dictionary[action] in ['south', 'north', 'east', 'west']:
                        moved, row, col = self._take_movement(index_action_dictionary[action], row, col)

                    # Check for collisions
                    if self.collision_sensitive_domain and moved:
                        if self.collided[taxi] == 0:
                            reward, moved, action, taxis_locations = self._check_action_for_collision(taxi,
                                                                                                      taxis_locations,
                                                                                                      row, col, moved,
                                                                                                      action, reward)

                    # Pickup
                    elif index_action_dictionary[action] == 'pickup':
                        passengers_status, reward = self._make_pickup(taxi, passengers_start_locations,
                                                                      passengers_status, taxi_location, reward)

                    # Dropoff
                    elif index_action_dictionary[action] == 'dropoff':
                        passengers_status, passengers_start_locations, reward = self._make_dropoff(taxi,
                                                                                                   passengers_start_locations,
                                                                                                   passengers_status,
                                                                                                   destinations,
                                                                                                   taxi_location,
                                                                                                   reward)

                    # Turning engine off
                    elif index_action_dictionary[action] == 'turn_engine_off':
                        reward = self.partial_closest_path_reward('turn_engine_off')
                        self.engine_status_list[taxi] = 0

                    # Standby with engine on
                    elif index_action_dictionary[action] == 'standby':
                        reward = self.partial_closest_path_reward('standby_engine_on')

                # Here we have finished checking for action for taxi-i
                # Fuel consumption
                if moved:
                    reward, fuels[taxi], taxis_locations = self._update_movement_wrt_fuel(taxi, taxis_locations,
                                                                                          row, col, reward, fuel)

                if (not moved) and action in [self.action_index_dictionary[direction] for
                                              direction in ['north', 'south', 'west', 'east']]:
                    reward = self.reward_dict['hit_wall']

                # taxi refuel
                if index_action_dictionary[action] == 'refuel':
                    reward, fuels[taxi] = self._refuel_taxi(fuel, reward, taxi, taxis_locations)

                # check if all the passengers are at their destinations
                done = all(loc == 1 for loc in passengers_status)
                self.dones[taxi_name] = done

                # check if all taxis collided
                done = all(self.collided == 1)
                self.dones[taxi_name] = self.dones[taxi_name] or done

                # check if all taxis are out of fuel
                done = fuels[taxi] == 0
                self.dones[taxi_name] = self.dones[taxi_name] or done

                rewards[taxi_name] = reward
                self.state = [taxis_locations, fuels, passengers_start_locations, destinations, passengers_status]
                self.last_action = action_dict

        self.dones['__all__'] = True
        self.dones['__all__'] = all(list(self.dones.values()))

        if self.bounded:
            total_reward = 0
            for taxi_id in action_dict.keys():
                total_reward += rewards[taxi_id]
            total_reward /= len(action_dict.keys())
            for taxi_id in action_dict.keys():
                rewards[taxi_id] = total_reward

        obs = {}
        for taxi_id in action_dict.keys():
            obs[taxi_id] = self.get_observation(self.state, taxi_id)
        obs, reward, done = obs[TAXI_NAME][0], {taxi_id: rewards[taxi_id] for taxi_id in action_dict.keys()}[
            TAXI_NAME], self.dones[TAXI_NAME]
        if self.taxi_loc_x_transform:
            obs[0] = 0
        if self.taxi_loc_y_transform:
            obs[1] = 0
        if self.pass_loc_x_transform:
            obs[3] = 4
        if self.pass_loc_y_transform:
            obs[4] = 0
        if self.pass_dest_x_transform:
            obs[5] = 4
        if self.pass_dest_y_transform:
            obs[6] = 4
        if self.pass_status_transform:
            obs[7] = 3
        obs = self.encode(obs)
        return obs, reward, done, {}

    def partial_closest_path_reward(self, basic_reward_str: str, taxi_index: int = None) -> int:
        """
        Computes the reward for a taxi and it's defined by:
        dropoff[s] - gets the reward equal to the closest path multiply by 15, if the drive got a passenger further
        away - negative.
        other actions - basic reward from config table
        Args:
            basic_reward_str: the reward we would like to give
            taxi_index: index of the specific taxi

        Returns: updated reward

        """
        if basic_reward_str not in ['intermediate_dropoff', 'final_dropoff'] or taxi_index is None:
            return self.reward_dict[basic_reward_str]

        # [taxis_locations, fuels, passengers_start_locations, destinations, passengers_status]
        current_state = self.state
        passengers_start_locations = current_state[2]

        taxis_locations = current_state[0]

        passengers_status = current_state[-1]
        passenger_index = passengers_status.index(taxi_index + 3)
        passenger_start_row, passenger_start_col = passengers_start_locations[passenger_index]
        taxi_current_row, taxi_currrent_col = taxis_locations[taxi_index]

        return 15 * (self.passenger_destination_l1_distance(passenger_index, passenger_start_row, passenger_start_col) -
                     self.passenger_destination_l1_distance(passenger_index, taxi_current_row, taxi_currrent_col))

    def get_explanation(self):
        explanation = []
        if self.fuel_transform:
            explanation.append(FUELS_TRANSFORM)
        # if self.status_transform:
        #     explanation.append(STATUS_TRANSFORM)
        if self.reward_transform:
            explanation.append(REWARD_TRANSFORM)
        return explanation

    def _take_movement(self, action: str, row: int, col: int) -> (bool, int, int):
        """
        Takes a movement with regard to a specific location of a taxi,
        can take action even though there a wall in the direction of the action.
        Args:
            action: direction to move
            row: current row
            col: current col

        Returns: if moved (always true), new row, new col

        """
        moved = False
        new_row, new_col = row, col
        max_row = self.num_rows - 1
        max_col = self.num_columns - 1
        if action == 'south':  # south
            if row != max_row:
                moved = True
            new_row = min(row + 1, max_row)
        elif action == 'north':  # north
            if row != 0:
                moved = True
            new_row = max(row - 1, 0)
        if action == 'east' and (self.no_walls_transform or self.desc[1 + row, 2 * col + 2] == b":"):  # east
            if col != max_col:
                moved = True
            new_col = min(col + 1, max_col)
        elif action == 'west' and (self.no_walls_transform or self.desc[1 + row, 2 * col] == b":"):  # west
            if col != 0:
                moved = True
            new_col = max(col - 1, 0)

        return moved, new_row, new_col

    #
    # TAXI_TRANSFORM_LIST = [TAXI_LOC_X, TAXI_LOC_Y, FUEL, PASS_LOC_X, PASS_LOC_Y, PASS_DEST_X, PASS_DEST_Y,
    # PASS_STATUS, WALLS, REWARD]


def get_taxi_transform_name(transforms):
    taxi_loc_x_transform, taxi_loc_y_transform = transforms[0], transforms[1]
    fuel_transform = transforms[2]
    pass_loc_x_transform, pass_loc_y_transform = transforms[3], transforms[4]
    pass_dest_x_transform, pass_dest_y_transform = transforms[5], transforms[6]
    pass_status_transform = transforms[7]
    no_walls_transform = transforms[8]
    reward_transform = transforms[9]
    name = ""
    name += TAXI_LOC_X if taxi_loc_x_transform else ""
    name += TAXI_LOC_Y if taxi_loc_y_transform else ""
    name += FUEL if fuel_transform else ""
    name += PASS_LOC_X if pass_loc_x_transform else ""
    name += PASS_LOC_Y if pass_loc_y_transform else ""
    name += PASS_DEST_X if pass_dest_x_transform else ""
    name += PASS_DEST_Y if pass_dest_y_transform else ""
    name += PASS_STATUS if pass_status_transform else ""
    name += WALLS if no_walls_transform else ""
    name += REWARD if reward_transform else ""
    return name



import copy
import pickle
import pandas as pd
import numpy as np
from itertools import combinations

# from Environments.SingleTaxiEnv.single_taxi_env import SingleTaxiEnv


def change_representation(state_info_i):
    for k, v in state_info_i.items():
        if len(v[0]) < 4:
            state_info_i[k] = [(0, -1, 0, False)]
    z = pd.DataFrame(state_info_i)
    y = z.to_numpy()[0]
    arr_2d = np.array([*y])
    new_states_vector = arr_2d[:, 1]
    return new_states_vector


def divide_states_for_allowed_and_not(P, decode, num_actions):
    # EACH INDEX IN THE LIST IS AN ACTION!
    allowed_states = [set() for _ in range(num_actions)]
    not_allowed_states = [set() for _ in range(num_actions)]
    for state, state_info in P.items():
        for action, action_info in state_info.items():
            action_info_i = refactor_to_dictionary(action_info)
            current_state = state
            current_state_f = decode(current_state)
            new_states_vector = change_representation(action_info_i)
            for s in new_states_vector:
                if s == -1:
                    continue
                state_f = np.array(decode(s))
                feature_diff = current_state_f - state_f
                # only to calc zero feature vec
                zero_state = tuple([0] * len(state_f))
                if tuple(feature_diff) != zero_state:
                    allowed_states[action].add(tuple(current_state_f))
                else:
                    not_allowed_states[action].add(tuple(current_state_f))
    return allowed_states, not_allowed_states


def get_data_structure(num_actions):
    tmp = {}
    for act in range(num_actions):
        tmp[act] = [set() for _ in range(num_actions)]
    return tmp


def refactor_to_dictionary(action_info):
    tmp = {}
    for k, v in enumerate(action_info):
        tmp[k] = [v]
    return tmp


def get_allowed_features(action_feature_sets):
    # f_names = ["new_row", "new_col", "new_pass_idx", "dest_idx", "fuel"]
    # a_names = ["SOUTH", "NORTH", "EAST", "WEST", "PICKUP", "DROPOFF", "REFUEL"]
    arr = np.array(list(action_feature_sets[0]))
    _, num_features = arr.shape
    pred_dict = {}
    for action in range(len(action_feature_sets)):
        pred_dict[action] = dict()
        action_info = action_feature_sets[action]
        arr = np.array(list(action_info))
        if not arr.any():
            continue
        feature_dict = dict()
        for feature_index in range(num_features):
            idx_list = list(combinations([_ for _ in range(num_features)], feature_index + 1))
            for idx in idx_list:
                idx_array = np.array([arr[:, k] for k in idx]).T
                f_pred = np.unique(idx_array, axis=0)
                feature_dict[idx] = f_pred
        pred_dict[action] = feature_dict
    return pred_dict


def get_not_allowed_features(action_feature_allowed, action_feature_not_allowed):
    _, num_features = np.array(list(action_feature_allowed[0])).shape
    pred_dict = {}
    for action in range(len(action_feature_allowed)):
        pred_dict[action] = dict()
        action_info_allowed = action_feature_allowed[action]
        action_info_not_allowed = action_feature_not_allowed[action]
        arr_allowed = np.array(list(action_info_allowed))
        arr_not_allowed = np.array(list(action_info_not_allowed))
        if not arr_allowed.any() or not arr_not_allowed.any():
            continue
        feature_dict = {}
        for feature_index in range(num_features):
            idx_list = list(combinations([_ for _ in range(num_features)], feature_index + 1))
            for idx in idx_list:
                idx_allowed = np.array([arr_allowed[:, k] for k in idx]).T
                idx_not_allowed = np.array([arr_not_allowed[:, k] for k in idx]).T
                f_pred_allowed = np.unique(idx_allowed, axis=0)
                f_pred_not_allowed = np.unique(idx_not_allowed, axis=0)
                real_not_allowed = []
                for not_allowed_pred in f_pred_not_allowed:
                    if len(np.where((f_pred_allowed == not_allowed_pred).all(axis=1))[0]) == 0:
                        real_not_allowed.append(not_allowed_pred)
                if len(real_not_allowed) != 0:
                    feature_dict[idx] = np.array(real_not_allowed)
        pred_dict[action] = feature_dict
    return pred_dict


def clean_pyramid(features_to_clean):
    features_to_clean = copy.deepcopy(features_to_clean)
    clean_idx = {}
    for action, action_info in features_to_clean.items():
        clean_idx[action] = dict()
        for i, (idx_i, idx_val_i) in enumerate(action_info.items()):
            for j, (idx_j, idx_val_j) in enumerate(action_info.items()):
                if i < j and len(idx_i) != len(idx_j):
                    same_idx = list(filter(lambda x: x in idx_i, list(idx_j)))
                    i_same_idx = np.where(np.array(idx_i) == same_idx)
                    j_same_idx = np.where(np.array(idx_j) == same_idx)
                    for val_i in idx_val_i:
                        for k, val_j in enumerate(idx_val_j):
                            is_same_idx = np.array([val_i[i_same_idx] == val_j[j_same_idx]]).all() if same_idx else False
                            if is_same_idx:
                                if idx_j not in clean_idx[action]:
                                    clean_idx[action][idx_j] = [k]
                                else:
                                    clean_idx[action][idx_j].append(k)

    for action, idx_list in clean_idx.items():
        for (idx_j, k) in idx_list.items():
            features_to_clean[action][idx_j] = np.delete(features_to_clean[action][idx_j], list(set(k)), axis=0)
            if not features_to_clean[action][idx_j].any():
                del features_to_clean[action][idx_j]
    return features_to_clean


class EnvPreconditions:
    def __init__(self, env):
        self.env = env
        self.allowed_states, self.not_allowed_states = divide_states_for_allowed_and_not(env.P, env.decode,
                                                                                         env.num_actions)
        # a_file = open("taxi_example_data/allowed_states.pkl", "rb")
        # self.allowed_states = pickle.load(a_file)
        # a_file = open("taxi_example_data/not_allowed_states.pkl", "rb")
        # self.not_allowed_states = pickle.load(a_file)

        self.allowed_features = get_allowed_features(self.allowed_states)
        # a_file = open("taxi_example_data/allowed_features.pkl", "rb")
        # self.allowed_features = pickle.load(a_file)

        not_allowed_features_with_duplicates = get_not_allowed_features(self.allowed_states, self.not_allowed_states)
        # a_file = open("taxi_example_data/not_allowed_features_with_duplicates.pkl", "rb")
        # not_allowed_features_with_duplicates = pickle.load(a_file)

        self.not_allowed_features = clean_pyramid(not_allowed_features_with_duplicates)
        # a_file = open("taxi_example_data/not_allowed_features.pkl", "rb")
        # self.not_allowed_features = pickle.load(a_file)

    def is_the_value_valid(self, action, idx, value_to_check):
        idx = (idx,) if isinstance(idx, int) else idx
        value_to_check = [value_to_check] if isinstance(value_to_check, int) else value_to_check
        if idx in self.allowed_features[action].keys():
            if value_to_check in self.allowed_features[action][idx]:
                return True
        return False

    def get_not_allowed_dictionary(self):
        return self.not_allowed_features

    def get_relevant_preconditions(self, anticipated_policy):
        relevant_preconditions = {}
        for state, action in anticipated_policy.items():
            relevant_preconditions[action] = {}
            if action in self.not_allowed_features.keys():
                relevant_preconditions[action] = self.not_allowed_features[action]
        return relevant_preconditions


# if __name__ == '__main__':
    # env_default_values = [0, 0, 0, 1, MAX_FUEL - 1]
    # state_visibility_indexes = []
    # cur_transforms = {STATE_VISIBILITY_TRANSFORM: (state_visibility_indexes, env_default_values),
    # ALL_OUTCOME_DETERMINIZATION: False,
    # MOST_LIKELY_OUTCOME: True}
    # env = SingleTaxiTransformedEnv(cur_transforms)
    # f_names = ["new_row", "new_col", "new_pass_idx", "dest_idx", "fuel"]
    # a_names = ["SOUTH", "NORTH", "EAST", "WEST", "PICKUP", "DROPOFF", "REFUEL"]

    # taxi_env = SingleTaxiEnv(deterministic=True)
    # preconditions = EnvPreconditions(taxi_env)
    # pass



from src.rlpe.experiments import *
from src.rlpe.utils import *


def temp_taxi_example_main():
    # define the environment
    env_name = SINGLE_TAXI_EXAMPLE
    agent_name = KERAS_DQN
    num_of_episodes = 200000
    results = {}
    original_env = get_env(env_name)

    # anticipated policy is <state, action> pairs
    anticipated_policy = ANTICIPATED_POLICY

    new_agent = load_existing_agent(original_env, agent_name, ORIGINAL_ENV, TRAINED_AGENTS_DIR_PATH)

    evaluation_result = rl_agent.run(new_agent, 1, method=EVALUATE, print_process=False, visualize=False)

    anticipated_policy_achieved, success_rate = is_anticipated_policy_achieved(original_env, new_agent,
                                                                               anticipated_policy)
    if anticipated_policy_achieved:
        print("The algorithm achieved the policy. We finished our work.")

    # transforms = load_existing_transforms_from_dir()
    transform_list = os.listdir(TRANSFORMS_PATH)
    # cur_transform_name = "0_(4,)_[0]_1_(4,)_[0]_2_(4,)_[0]_4_(4,)_[0]_5_(4,)_[0].pkl"
    # t_name, t_env = load_transform_by_name(cur_transform_name, dir_name=TRANSFORMS_PATH)
    # transforms = {0: (t_name, t_env)}
    explanations = []
    for f in transform_list:
        # for params, (transform_name, transformed_env) in transforms.items():
        # create agent
        transform_name, transformed_env = load_transform_by_name(f)
        print(f"\nEvaluating agent on the transformed environment: {transform_name}")
        agent = load_existing_agent(transformed_env, agent_name, transform_name, TRAINED_AGENTS_DIR_PATH)
        if agent is None:
            continue
        transformed_train_result, explanation = load_existing_results(agent_name, transform_name, num_of_episodes)
        # evaluate the performance of the agent
        transformed_evaluation_result = rl_agent.run(agent, num_of_episodes, method=EVALUATE, print_process=True,
                                                     visualize=False)

        # check if the anticipated policy is achieved in trans_env
        anticipated_policy_achieved, success_rate = is_anticipated_policy_achieved(original_env, agent,
                                                                                   anticipated_policy)

        results[transform_name] = {EVALUATION_RESULTS: transformed_evaluation_result,
                                   TRAINING_RESULTS: transformed_train_result,
                                   SUCCESS_RATE: success_rate,
                                   GOT_AN_EXPLANATION: False}
        if anticipated_policy_achieved:
            print(f"Got an explanation: {transform_name}")
            explanations.append(transform_name)
            results[transform_name][GOT_AN_EXPLANATION] = True

    if explanations is None or len(explanations) == 0:
        print("No explanation found! :-(")
    else:
        print(f"Explanations found: {explanations}")

    success_rates = [v[SUCCESS_RATE] for v in results.values()]
    names = [_ for _ in range(len(success_rates))]
    names_translation_dict = dict()
    for i, k in enumerate(results.keys()):
        names_translation_dict[i] = k
    fig, ax = plt.subplots()
    ax.bar(names, success_rates)
    ax.set_xticks(np.arange(len(names)))
    ax.set_ylabel("Success Rate")
    ax.set_title("Success Rate of the different transformed environments")
    print(f"for translating the labels into transform name: \n{names_translation_dict}")
    plt.show()
    a = 7


def colab_notebook_example():
    original_env = get_env(env_name=SINGLE_TAXI_EXAMPLE)
    actor = load_existing_agent(env=original_env, agent_name=KERAS_DQN, env_name=ORIGINAL_ENV,
                                trained_agents_path=TRAINED_AGENTS_DIR_PATH)
    rl_agent.run(actor, 1, method=EVALUATE, print_process=True, visualize=True)



import numpy as np

from keras.models import Sequential
from keras.layers import Dense, Activation, Flatten
# from keras.optimizers import Adam

from rl.agents.cem import CEMAgent
from rl.memory import EpisodeParameterMemory

from src.rlpe.constants import *
from src.rlpe.Agents.abstract_agent import AbstractAgent


class KerasCEMAgent(AbstractAgent):
    def __init__(self, env, timesteps_per_episode=10001):
        super().__init__(env, timesteps_per_episode)
        self.evaluating = False
        self.action_size = env.action_space.n
        self.state_size = env.num_states
        self.model = self._build_compile_model()
        memory = EpisodeParameterMemory(limit=1000, window_length=1)
        self.agent = CEMAgent(model=self.model, nb_actions=self.action_size, memory=memory, batch_size=50,
                              nb_steps_warmup=2000, train_interval=50, elite_frac=0.05)

    def run(self) -> {str: float}:
        """
        The agent's training method.
        Returns: a dictionary - {"episode_reward_mean": __, "episode_reward_min": __, "episode_reward_max": __,
        "episode_len_mean": __}
        """
        self.agent.compile()
        history = self.agent.fit(self.env, nb_steps=ITER_NUM, visualize=False, verbose=1)
        if len(history.history) > 0:
            episode_reward = history.history["episode_reward"]
            nb_episode_steps = history.history["nb_episode_steps"]
        else:
            episode_reward, nb_episode_steps = [0], [0]  # TODO - placeholder
        result = {EPISODE_REWARD_MEAN: np.array(episode_reward),
                  EPISODE_STEP_NUM_MEAN: np.array(nb_episode_steps),
                  EPISODE_REWARD_MIN: np.empty([]),
                  EPISODE_REWARD_MAX: np.empty([]), EPISODE_VARIANCE: np.empty([])}
        return result

    def _build_compile_model(self):
        ## simple net
        # model = Sequential()
        # model.add(Flatten(input_shape=(1,) + self.env.observation_space.shape))
        # model.add(Dense(self.action_size))
        # model.add(Activation('softmax'))
        # return model
        model = Sequential()
        model.add(Flatten(input_shape=(1,) + self.env.observation_space.shape))
        model.add(Dense(16))
        model.add(Activation('relu'))
        model.add(Dense(16))
        model.add(Activation('relu'))
        model.add(Dense(16))
        model.add(Activation('relu'))
        model.add(Dense(self.action_size))
        model.add(Activation('softmax'))
        return model

    def compute_action(self, state) -> int:
        """
        Computes the best action from a given state.
        Returns: a int that represents the best action.
        """
        # state = np.array([[state]])
        return int(np.argmax(self.model.predict(state)))

    def stop_episode(self):
        pass

    def episode_callback(self, state, action, reward, next_state, terminated):
        pass

    def evaluate(self, visualize=False):
        self.agent.test(self.env, nb_episodes=5, visualize=visualize, nb_max_episode_steps=60)

    def replay_experiences(self):
        pass



from src.rlpe.Environments.SingleTaxiEnv.single_taxi_env import *

SMALL_EXAMPLE_MAP = [
    "+-----+",
    "|R| :G|",
    "| : : |",
    "|Y:F|B|",
    "+-----+",
]


class SingleSmallTaxiEnv(SingleTaxiEnv):
    """
    The Taxi Problem
    from "Hierarchical Reinforcement Learning with the MAXQ Value Function Decomposition"
    by Tom Dietterich
    Description:
    There are four designated locations in the grid world indicated by R(ed), G(reen), Y(ellow), and B(lue). When the episode starts, the taxi starts off at a random square and the passenger is at a random location. The taxi drives to the passenger's location, picks up the passenger, drives to the passenger's destination (another one of the four specified locations), and then drops off the passenger. Once the passenger is dropped off, the episode ends.
    Observations:
    There are 500 discrete states since there are 25 taxi positions, 5 possible locations of the passenger (including the case when the passenger is in the taxi), and 4 destination locations.
    Passenger locations:
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)
    - 4: in taxi
    Destinations:
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)
    Actions:
    There are 6 discrete deterministic actions:
    - 0: move south
    - 1: move north
    - 2: move east
    - 3: move west
    - 4: pickup passenger
    - 5: drop off passenger
    - 6: refuel the taxi
    Rewards:
    There is a default per-step reward of -1,
    except for delivering the passenger, which is +20,
    or executing "pickup" and "drop-off" actions illegally, which is -10.
    Rendering:
    - blue: passenger
    - magenta: destination
    - yellow: empty taxi
    - green: full taxi
    - other letters (R, G, Y and B): locations for passengers and destinations
    state space is represented by:
        (taxi_row, taxi_col, passenger_location, destination)
    """
    metadata = {'render.modes': ['human', 'ansi']}

    def __init__(self, deterministic=True):
        self.init_state, self.init_row, self.init_col = 6152, 2, 0
        self.deterministic = deterministic
        self.desc = np.asarray(SMALL_EXAMPLE_MAP, dtype='c')
        w, h = self.desc.shape
        self.last_action = None
        self.passengers_locations, self.fuel_station = self.get_info_from_map()
        self.taxi_fuel = MAX_FUEL
        self.num_rows = int(w - 2)
        self.num_columns = int((h - 1) / 2)
        self.num_states = (self.num_rows * self.num_columns) * (len(self.passengers_locations) + 1) * (
            len(self.passengers_locations)) * MAX_FUEL
        self.max_row = self.num_rows - 1
        self.max_col = self.num_columns - 1
        self.initial_state_distribution = np.zeros(self.num_states)
        self.num_actions = len(ACTIONS)
        self.passenger_in_taxi = len(self.passengers_locations)
        self.P = self.build_transition_matrix(deterministic=deterministic)
        # self.initial_state_distribution /= self.initial_state_distribution.sum()
        discrete.DiscreteEnv.__init__(self, self.num_states, self.num_actions, self.P, self.initial_state_distribution)



import matplotlib.pyplot as plt
import numpy as np
from src.rlpe.constants import *

# mpl_style(dark=False, minor_ticks=True)
figure_size = (12, 6)
legend_alpha = 0.5
SAVE_RATE = 500


def prepare_result_graph(result, sub_plot_name="", x_label="", y_label=""):
    plt.plot(range(len(result)), result, label=sub_plot_name)
    plt.xlabel(x_label)
    plt.ylabel(y_label)


def plot_result_graphs(agent_name, result):
    names = []
    train_episode_reward_mean = []
    evaluate_episode_reward_mean = []
    success_rate = []
    for name, res in result.items():
        names.append(name)
        train_episode_reward_mean.append(res["train_episode_reward_mean"])
        evaluate_episode_reward_mean.append(res["evaluate_episode_reward_mean"])
        success_rate.append(res["success_rate"])

    plot_result_graph(agent_name, train_episode_reward_mean, names, "train_episode_reward_mean")
    plot_result_graph(agent_name, evaluate_episode_reward_mean, names, "evaluate_episode_reward_mean")
    plot_success_rate_charts(names, success_rate, 'success_rate by transforms', 'success_rate')


def plot_graph_by_transform_name_and_env(agent_name, result, title=""):
    for transform_name, reward in result.items():
        prepare_result_graph(reward, sub_plot_name=transform_name, x_label="environments", y_label="mean_reward")
    show_fig(agent_name, title)


def plot_result_graph(agent_name, results, names, title):
    for res, name in zip(results, names):
        prepare_result_graph(res, sub_plot_name=name, x_label='episodes', y_label='episode_reward_mean')
    show_fig(agent_name, title)


def show_fig(agent_name, title):
    plot_name = agent_name + "_" + title
    plt.title(plot_name)
    plt.legend()
    plt.savefig(plot_name + '.png')
    plt.show()


def plot_all_success_rate_charts(result, save_folder, file_name="", save_fig=False):
    all_success_rate = {}
    for res in result:
        for name, res_by_name in res.items():
            if name not in all_success_rate:
                all_success_rate[name] = [res_by_name[SUCCESS_RATE]]
            else:
                all_success_rate[name].append(res_by_name[SUCCESS_RATE])
    all_success_rate_mean = {}
    for name, res in all_success_rate.items():
        success_rate = np.array(res)
        success_rate_mean = success_rate.mean()
        all_success_rate_mean[name] = success_rate_mean
    names = list(all_success_rate_mean.keys())
    success_rate_means = list(all_success_rate_mean.values())
    plot_success_rate_charts(names, success_rate_means, plot_name="chart_graph_" + file_name, save_folder=save_folder,
                             save_fig=save_fig)
    # for i in range(32):
    #     plot_success_rate_charts(names[i * 32: (i + 1) * 32], success_rate_means[i * 32: (i + 1) * 32],
    #                              plot_name=str(i) + "_chart_graph", save_folder=save_folder)


def plot_success_rate_charts(names, success_rate, plot_name="", save_folder=None, y_label="", save_fig=False):
    fig, ax = plt.subplots()
    ax.bar(names, success_rate)
    ax.set_ylabel(y_label)
    ax.set_title(plot_name)
    for tick in ax.get_xticklabels():
        tick.set_rotation(90)
    fig.align_labels()
    if save_fig:
        plt.savefig('output/' + plot_name + '.png')
    plt.show()


def get_plot(plt, x, stats, name, color, which_results, line="-"):
    stats_ = stats
    mean, std, q1, q3 = process_stats(stats_, name, which_results)
    mean, std, q1, q3 = mean[:len(x)], std[:len(x)], q1[:len(x)], q3[:len(x)]
    # mean_smoothed = pd.Series(mean).rolling(5, min_periods=5).mean()
    cumsum = type == 'cumsum'
    if cumsum:
        plt.plot(x, np.cumsum(mean), line, color=color)
    else:
        plt.plot(x, mean, line, color=color)
        plt.fill_between(x, q1, q3, color=color, alpha=0.1)


def process_stats(result, name, which_results, reward=True):
    '''
    returns: episode_lengths, episode_rewards, None, None or
             mean_episode_lengths, mean_episode_rewards, std_episode_lengths, std_episode_rewards
    '''

    if len(result) == 1:
        return result[0][name][which_results][EPISODE_STEP_NUM_MEAN], result[0][name][which_results][
            EPISODE_REWARD_MEAN], [0], [0]
    elif len(result) > 1:
        all_episode_lengths = []
        all_episode_variance = []
        all_episode_rewards = []
        all_episode_rewards = []

        for s in result:
            all_episode_lengths.append(s[name][which_results][EPISODE_STEP_NUM_MEAN])
            all_episode_rewards.append(s[name][which_results][EPISODE_REWARD_MEAN])

        all_episode_lengths = np.array(all_episode_lengths)
        all_episode_rewards = np.array(all_episode_rewards)

        if reward:
            q1 = np.percentile(all_episode_rewards, 25, axis=0, interpolation='midpoint')
            q3 = np.percentile(all_episode_rewards, 75, axis=0, interpolation='midpoint')
            return all_episode_rewards.mean(axis=0), all_episode_rewards.std(axis=0), q1, q3

        else:
            q1 = np.percentile(all_episode_lengths, 25, axis=0, interpolation='midpoint')
            q3 = np.percentile(all_episode_lengths, 75, axis=0, interpolation='midpoint')
            return all_episode_lengths.mean(axis=0), all_episode_lengths.std(axis=0), q1, q3


def plot_reward_graph(result, which_results, save_folder, file_name="", save_fig=False):
    max_length = len(result[0][ORIGINAL_ENV][TRAINING_RESULTS][EPISODE_REWARD_MEAN])
    epochs_num = len(result)
    fig1, ax = plt.subplots(figsize=figure_size)
    x = np.arange(1, max_length + 1)
    transforms = []
    for transform_name, res in result[0].items():
        get_plot(plt, x, result, transform_name, None, which_results)
        transforms.append(transform_name)
    plt.legend(transforms, framealpha=legend_alpha)
    ax.set_xlim(xmin=0)
    ax.set_xlim(xmax=max_length)
    plt.xlabel("Episode")
    plt.ylabel("Episode Reward")
    plt.title(which_results)
    if save_fig:
        plt.savefig('output/' + file_name + EPISODE_REWARD_MEAN + '.png')
    plt.show()


def plot_results(result, save_folder, file_name="", save_fig=False):
    plot_reward_graph(result, TRAINING_RESULTS, save_folder, file_name=TRAINING_RESULTS + file_name, save_fig=save_fig)
    plot_reward_graph(result, EVALUATION_RESULTS, save_folder, file_name=EVALUATION_RESULTS + file_name,
                      save_fig=save_fig)
    plot_all_success_rate_charts(result, save_folder, file_name=file_name, save_fig=save_fig)


def prepare_calc_mean(val_dict):
    for result in val_dict[EVALUATION_RESULTS].keys():
        val_dict[EVALUATION_RESULTS][result] = calc_mean(val_dict[EVALUATION_RESULTS][result])
    for result in val_dict[TRAINING_RESULTS].keys():
        val_dict[TRAINING_RESULTS][result] = calc_mean(val_dict[TRAINING_RESULTS][result])
    return val_dict


def calc_mean(val_list):
    mean_lst = []
    cum_res = 0
    i = 0
    for val in val_list:
        if i == SAVE_RATE:
            mean_lst.append(cum_res / SAVE_RATE)
            i, cum_res = 0, 0
        else:
            cum_res += val
            i += 1
    return mean_lst



class Agent():

    # init agents and their observations
    def __init__(self, decision_maker, observation):
        self.decision_maker = decision_maker
        self.observation = observation

    def get_action(self, environment, cur_observation):
        return self.decision_maker.get_decision(environment, cur_observation)

    def get_observation(self, state, param):
        return self.observation(state, param)


from itertools import product

from src.rlpe.Environments.abstract_wrapper_env import AbstractWrapperEnv


class SingleTaxiSimpleEnv(SingleTaxiEnv, AbstractWrapperEnv):
    def __init__(self, deterministic=True):
        super().__init__(deterministic)
        self.transform_num = 7

    def is_possible_initial_state(self, pass_idx, dest_idx, row, col):  # Fix taxi initial position
        return pass_idx < 4 and pass_idx != dest_idx and row == self.init_row and col == self.init_col

    def get_states_from_partial_obs(self, partial_obs):
        partial_obs_aligned_with_env = False
        iter_num = 200
        while not partial_obs_aligned_with_env and iter_num != 0:
            obs = self.decode(self.reset())
            if self._is_aligned(obs, partial_obs):
                partial_obs_aligned_with_env = True
            iter_num -= 1

        if partial_obs_aligned_with_env:
            taxi_x = [partial_obs[0]] if (partial_obs[0] is not None) else list(range(self.num_columns))
            taxi_y = [partial_obs[1]] if (partial_obs[1] is not None) else list(range(self.num_rows))
            passenger_loc = [partial_obs[2]] if (partial_obs[2] is not None) else list(
                range(len(self.passengers_locations)))
            passenger_dest = [partial_obs[3]] if (partial_obs[3] is not None) else list(
                range(len(self.passengers_locations)))
            fuel = [partial_obs[4]] if partial_obs[4] else list(range(0, MAX_FUEL))
            states = list(product(taxi_x, taxi_y, passenger_loc, passenger_dest, fuel, repeat=1))
            states = [self.encode(*state) for state in states]
        else:
            states = []
        return states

    def _is_aligned(self, obs, partial_obs):
        #  taxi_x, taxi_y, passenger_loc, passenger_dest, fuel
        taxi_x, taxi_y, passenger_loc, passenger_dest, fuel = obs
        return passenger_dest == partial_obs[3]



from src.rlpe.Transforms.transform_constants import *
from src.rlpe.experiments import *


def run_single_taxi_env():
    env_name = SINGLE_TAXI_EXAMPLE
    agent_name = KERAS_DQN
    num_states_in_partial_policy = 5
    num_of_epochs = 1
    num_of_episodes_per_epoch = 10000
    default_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch, num_states_in_partial_policy)


def run_taxi_env():
    env_name = TAXI_EXAMPLE
    agent_name = KERAS_DQN
    num_states_in_partial_policy = 5
    num_of_epochs = 1
    num_of_episodes_per_epoch = 100
    different_anticipated_policy_size_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch)


def run_lunar_lander_env():
    env_name = LUNAR_LANDER
    agent_name = KERAS_DQN
    num_states_in_partial_policy = 5
    num_of_epochs = 5
    num_of_episodes_per_epoch = 500
    default_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch, num_states_in_partial_policy)


def run_search_transform_taxi_env_example_colab():
    env_name = SEARCH_TRANSFORM_TAXI_ENV
    anticipated_policy = ANTICIPATED_POLICY
    num_of_episodes_per_epoch = ITER_NUM
    agent_name = KERAS_DQN

    original_env = get_env(SINGLE_TAXI_EXAMPLE)
    search_taxi_env = get_env(env_name)
    agent = load_existing_agent(search_taxi_env, agent_name, env_name)
    result = {}

    # evaluate the performance of the agent
    transformed_evaluation_result = rl_agent.run(agent, num_of_episodes_per_epoch, method=EVALUATE)

    # check if the anticipated policy is achieved in trans_env
    anticipated_policy_achieved, success_rate = is_anticipated_policy_achieved(original_env, agent, anticipated_policy,
                                                                               search_taxi_env)

    result[env_name] = load_pkl_file(TRAINED_AGENT_RESULT_FILE_PATH)
    if anticipated_policy_achieved:
        result[env_name][GOT_AN_EXPLANATION] = True

    if result[env_name][GOT_AN_EXPLANATION]:
        print(f"\nexplanation found! on the {env_name} environment.")
        explanation = map_actions_to_explanation(original_env, agent, search_taxi_env, anticipated_policy)
        for transform_name, actions in explanation.items():
            print(f"transform name: {transform_name} for mapping action {actions[0]} to {actions[1]}")
        print(explanation)
    else:
        print("no explanation found:-(")


if __name__ == '__main__':
    # preconditions = load_pkl_file(PRECONDITIONS_PATH)
    # agent_name = KERAS_CEM
    # env_name = TAXI_EXAMPLE
    # new_env = get_env(env_name)
    #
    # agent, restored = make_or_restore_model(new_env, agent_name, env_name)
    #
    # print(f"\nTraining and evaluating the {agent_name} on \"{env_name}\" environment")
    # train_result = rl_agent.run(agent, 1, method=TRAIN)
    # rl_agent.run(agent, 5, method=EVALUATE)
    # import csv
    # frozen_lake_satisfaction_rates = []
    # with open('results/satisfaction_rate_frozen_lake.csv', newline='') as csvfile:
    #     spamreader = csv.reader(csvfile, delimiter=' ', quotechar='|')
    #     for row in spamreader:
    #         frozen_lake_satisfaction_rates.append(float(row[1]))
    #         print(row)
    from src.rlpe.create_single_taxi_transforms import *
    success_rates = check_success_rate()

    print("DONE!")



from abc import ABC, abstractmethod


class AbstractEnvironment(ABC):
    """
    This is the basic environment requirements for working with this project.
    """
    @abstractmethod
    def reset(self):
        """
          Resets the current state to the start state
        """
        pass

    @abstractmethod
    def render(self):
        """
        Renders the environment.
        """
        pass

    @abstractmethod
    def step(self, action):
        """
          Performs the given action in the current
          environment state and updates the environment.

          Returns (new_obs, reward, done, info)
        """
        pass



import multiprocessing
from src.rlpe.visualize import *
from src.rlpe.Agents.RL_agents.q_learning_agents import *
import tensorflow as tf
from tensorflow.python.client import device_lib
from tensorflow.python.keras import backend as K
from src.rlpe.TransformSearch.anticipation_BFS import *
from src.rlpe.Transforms.env_precinditions import *


def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return len([x.name for x in local_device_protos if x.device_type == 'GPU'])


def log(string):
    print(string)
    f = open('output_log.log', 'wb')
    f.write(string)
    f.close()


def create_run_and_evaluate_agent(original_env, transformed_env, agent_name, env_name, num_of_episodes,
                                  anticipated_policy, result, explanation):
    # GPUtil.showUtilization()
    agent, restored = make_or_restore_model(transformed_env, agent_name, env_name)

    print(f"\nTraining and evaluating the {agent_name} on \"{env_name}\" environment")
    if not restored:
        train_result = rl_agent.run(agent, num_of_episodes, method=TRAIN)
    else:
        train_result = None

    evaluation_result = rl_agent.run(agent, num_of_episodes, method=EVALUATE)

    anticipated_policy_achieved, success_rate = is_anticipated_policy_achieved(original_env, agent, anticipated_policy)

    result[env_name] = {EVALUATION_RESULTS: evaluation_result,
                        TRAINING_RESULTS: train_result,
                        SUCCESS_RATE: success_rate,
                        GOT_AN_EXPLANATION: False}

    if anticipated_policy_achieved:
        print(f"Environment {env_name} achieved an explanation!")
        explanation.append(env_name)
        result[env_name][GOT_AN_EXPLANATION] = True

    save_trained_model(agent, agent_name, env_name)
    return result, explanation


def run_multiprocess(original_env, transformed_env, agent_name, transform_name, num_of_episodes, anticipated_policy,
                     explanation):
    process_list = []
    NUM_GPUS = get_available_gpus()
    manager = multiprocessing.Manager()
    result = manager.dict()  # create shared data structure
    num_process_running = 0
    if num_process_running < NUM_GPUS:
        num_process_running += 1
        p = multiprocessing.Process(target=create_run_and_evaluate_agent, args=(
            original_env, transformed_env, agent_name, transform_name, num_of_episodes,
            anticipated_policy, result, explanation))
        process_list.append(p)
        p.start()
    else:
        for process in process_list:
            process.join()

        process_list = []
        num_process_running = 1
        p = multiprocessing.Process(target=create_run_and_evaluate_agent, args=(
            original_env, transformed_env, agent_name, transform_name, num_of_episodes, anticipated_policy, result,
            explanation))
        process_list.append(p)
        p.start()

    for name, res in result.items():
        if res["anticipated_policy_achieved"]:
            explanation.append(name)
            real_name = name
            result[real_name][GOT_AN_EXPLANATION] = True
    return result, explanation


def run_experiment(env_name, agent_name, num_of_episodes, num_states_in_partial_policy):
    config = tf.compat.v1.ConfigProto()
    config.gpu_options.allow_growth = True
    session = tf.compat.v1.Session(config=config)
    K.set_session(session)

    result = {}
    original_env = get_env(env_name)

    # the anticipated policy (which is part of our input and defined by the user)
    # full_expert_policy_dict = get_expert(env_name, original_env).full_expert_policy_dict()
    # anticipated_policy = sample_anticipated_policy(full_expert_policy_dict, num_states_in_partial_policy)

    anticipated_policy = dict()
    anticipated_policy[(2, 0, 0, 3, None)] = [1]
    anticipated_policy[(1, 0, 0, 3, None)] = [1]
    anticipated_policy[(0, 0, 0, 3, None)] = [4]

    explanation = []

    # result, explanation = create_run_and_evaluate_agent(original_env, original_env, agent_name, ORIGINAL_ENV,
    #                                                     num_of_episodes,
    #                                                     anticipated_policy, result, explanation)
    transforms = dict()
    file_names = ['1_(4,)_[0]_4_(1,)_[2]_4_(0,)_[2].pkl', '4_(0,)_[2].pkl']
    for i, file_name in enumerate(file_names):
        transforms[i] = load_transform_by_name(file_name)
    # transforms = load_existing_transforms(env_name, anticipated_policy)

    for params, (transform_name, transformed_env) in transforms.items():
        result, explanation = run_multiprocess(original_env, transformed_env, agent_name, transform_name,
                                               num_of_episodes, anticipated_policy, explanation)
        # result, explanation = create_run_and_evaluate_agent(original_env, transformed_env, agent_name, transform_name,
        #                                                     num_of_episodes,
        #                                                     anticipated_policy, result, explanation)

    if explanation is None:
        print("no explanation found - you are too dumb for our system")
    else:
        print("explanation found %s:" % explanation)

    return result


def different_anticipated_policy_size_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch):
    # num_states = [1, 5, 10, 20, 30]
    num_states = [5]
    for num_states_in_partial_policy in num_states:
        default_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch, num_states_in_partial_policy)


def plot_results_by_number_of_transforms(pkl_name, output_folder, num_episodes):
    with open(pkl_name, 'rb') as f:
        data = pickle.load(f)
        result_1, result_2, result_3, result_4, result_5 = {}, {}, {}, {}, {}
        ls1, ls2, ls3, ls4, ls5 = [], [], [], [], []
        for cur_dict in data:
            for k, v in cur_dict.items():
                name = k.split('_')[:-1]
                if k == ORIGINAL_ENV:
                    orig_res = prepare_calc_mean(v)
                    result_1[k] = orig_res
                    result_2[k] = orig_res
                    result_3[k] = orig_res
                    result_4[k] = orig_res
                    result_5[k] = orig_res
                elif len(name) == 1:
                    result_1[k] = prepare_calc_mean(v)
                elif len(name) == 2:
                    result_2[k] = prepare_calc_mean(v)
                elif len(name) == 3:
                    result_3[k] = prepare_calc_mean(v)
                elif len(name) == 4:
                    result_4[k] = prepare_calc_mean(v)
                elif len(name) == 5:
                    result_5[k] = prepare_calc_mean(v)
            ls1.append(result_1)
            ls2.append(result_2)
            ls3.append(result_3)
            ls4.append(result_4)
            ls5.append(result_5)
            result_1, result_2, result_3, result_4, result_5 = {}, {}, {}, {}, {}
        save_cur_fig = False
        plot_results(ls1, output_folder, file_name="_1trans_" + str(num_episodes) + "_", save_fig=save_cur_fig)
        plot_results(ls2, output_folder, file_name="_2trans_" + str(num_episodes) + "_", save_fig=save_cur_fig)
        plot_results(ls3, output_folder, file_name="_3trans_" + str(num_episodes) + "_", save_fig=save_cur_fig)
        plot_results(ls4, output_folder, file_name="_4trans_" + str(num_episodes) + "_", save_fig=save_cur_fig)
        plot_results(ls5, output_folder, file_name="_5trans_" + str(num_episodes) + "_", save_fig=save_cur_fig)


def default_experiment(agent_name, env_name, num_of_epochs, num_of_episodes_per_epoch, num_states_in_partial_policy):
    # output_folder = "./output/"
    # if os.path.isdir(output_folder):
    #     shutil.rmtree(output_folder)
    # os.makedirs(output_folder)

    result = []

    for i in range(num_of_epochs):
        print('\nEpoch num:', i + 1)
        episode_result = run_experiment(env_name, agent_name, num_of_episodes_per_epoch, num_states_in_partial_policy)
        result.append(episode_result)

    # save result
    pkl_name = DATA_FOLDER + agent_name + '_all_stats_' + str(num_of_episodes_per_epoch) + "_states_" + str(
        num_states_in_partial_policy) + ".pkl"
    output = open(pkl_name, 'wb')
    pickle.dump(result, output)
    output.close()

    # plot
    # plot_results(result, output_folder)
    # plot_results_by_number_of_transforms(pkl_name, output_folder, num_of_episodes_per_epoch)
    return result


def different_envs_experiment():
    env_name = TAXI_EXAMPLE
    agent_name = Q_LEARNING
    num_of_episodes_per_epoch = 1000
    num_of_envs = 1
    num_states_in_partial_policy = 10
    all_env_results = {}
    all_env_test_results = {"original": [], "walls": [], "reward": [], "reward_walls": [], "fuel": [],
                            "fuel_walls": [], "fuel_reward": [], "fuel_reward_walls": []}
    all_env_evaluate_results = {"original": [], "walls": [], "reward": [], "reward_walls": [], "fuel": [],
                                "fuel_walls": [], "fuel_reward": [], "fuel_reward_walls": []}
    all_env_success_results = {"original": [], "walls": [], "reward": [], "reward_walls": [], "fuel": [],
                               "fuel_walls": [], "fuel_reward": [], "fuel_reward_walls": []}

    for i in range(num_of_envs):
        # set_up_env_idx()
        result = run_experiment(env_name, agent_name, num_of_episodes_per_epoch, num_states_in_partial_policy)
        all_env_results[i] = result
        for k, v in all_env_test_results.items():
            all_env_test_results[k].append(np.mean(result[k]['train_episode_reward_mean']))
            all_env_evaluate_results[k].append(np.mean(result[k]['evaluate_episode_reward_mean']))
            all_env_success_results[k].append(np.mean(result[k]['success_rate']))
        plot_result_graphs(agent_name, result)

        # visualize rewards and success_rate
    # names = ["original", "walls", "reward", "reward_walls", "fuel", "fuel_walls", "fuel_reward", "fuel_reward_walls"]
    # plot_graph_by_transform_name_and_env(agent_name, all_env_test_results, "300_env_test_results_5000_episodes")
    # plot_graph_by_transform_name_and_env(agent_name, all_env_evaluate_results, "300_env_evaluate_results_5000_episodes")
    # plot_success_rate_charts(names, [np.mean(v) for v in all_env_success_results.values()],
    #                          "success_rate_300_env_5000_episodes")


def run_anticipated_dfs_search():
    cur_env_preconditions = load_pkl_file(PRECONDITIONS_PATH)
    precondition_graph = load_pkl_file(PRECONDITION_GRAPH_PATH)
    if not precondition_graph:
        precondition_graph = PreconditionsGraph(SINGLE_TAXI_EXAMPLE, cur_env_preconditions.not_allowed_features,
                                                ANTICIPATED_POLICY)
        save_pkl_file("precondition_graph_small_taxi_env.pkl", precondition_graph)
    precondition_graph.bfs()


def create_transform_search_env():
    from src.rlpe.TransformSearch.taxi_transform_search_env import create_transform_search_taxi_env, reformat_preconditions

    env = get_env(SINGLE_TAXI_EXAMPLE)
    preconditions = load_pkl_file(PRECONDITIONS_PATH)
    new_env_preconditions = reformat_preconditions(preconditions)
    new_env = create_transform_search_taxi_env(env, new_env_preconditions, ANTICIPATED_POLICY)
    return new_env


# if __name__ == '__main__':
#     from TransformSearch.greedy_search import greedy_search
#
#     preconditions = load_pkl_file(PRECONDITIONS_PATH)
#     max_transformed_env_opt, best_cluster = greedy_search(preconditions, ANTICIPATED_POLICY)
#     a = 7


import collections
import numpy as np
from src.rlpe.save_load_utils import *


def preconditions_relaxation(preconditions_info, env):
    # preconditions = EnvPreconditions(env)
    a_file = open(PRECONDITIONS_PATH, "rb")
    preconditions = pickle.load(a_file)

    diff_dict, state_by_diff, next_state_by_action = get_diff_for_actions(env.P, env)
    # a_file = open("taxi_example_data/diff_dict.pkl", "rb")
    # diff_dict = pickle.load(a_file)

    for act, act_info in preconditions_info.items():
        if env.deterministic:  # deterministic case
            pre_process_info_and_update_p(env, act, diff_dict, act_info, act, preconditions, None)
        else:
            for act_prob, pre_info in act_info.items():
                pre_process_info_and_update_p(env, act_prob, diff_dict, pre_info, act, preconditions, act_prob)
    return preconditions


def pre_process_info_and_update_p(env, act_to_diff, diff_dict, pre_info, act, preconditions, act_prob=None):
    state_to_replace = get_state_to_replace(env, pre_info, act, preconditions)
    mapping_states_dict = get_mapping_states_dict(env, state_to_replace, diff_dict[act_to_diff])
    update_p_matrix_by_relax_preconditions(env, mapping_states_dict, act, act_prob)


def get_state_to_replace(env, preconditions_info, act, preconditions):
    state_to_replace = []
    for state in env.P.keys():
        decoded_state = np.array(env.decode(state))
        for idx, values in preconditions_info.items():
            for v in values:
                state_values = decoded_state[np.array([idx])]
                sufficient_for_action = is_sufficient_for_action(idx, v, preconditions.not_allowed_features, act,
                                                                 decoded_state)
                if np.array([state_values == np.array(v)]).all() and sufficient_for_action:
                    state_to_replace.append(state)
    return state_to_replace


def is_sufficient_for_action(relaxed_idx, relaxed_val, not_allowed_features, act, decoded_state):
    result = True
    for key, values in not_allowed_features[act].items():
        for val in values:
            if np.array([decoded_state[list(key)] == val]).all() and (
                    np.array([relaxed_idx != key]).all() or np.array([relaxed_val != val]).all()):
                result = False
                break
        if not result:
            break
    return result


def is_state_need_a_replacement(act, preconditions, preconditions_info, act_prob):
    if act_prob:
        allowed_features_by_action = preconditions.allowed_features[act][act_prob]
    else:
        allowed_features_by_action = preconditions.allowed_features[act]


def get_mapping_states_dict(env, state_to_replace, optional_diff_vec):
    mapping_states_dict = dict((s, 0) for s in state_to_replace)
    for state in state_to_replace:
        decoded_state = np.array(env.decode(state))
        diff_vec = np.array([0] * len(decoded_state))
        if len(optional_diff_vec) > 3:
            for key, val in optional_diff_vec.items():
                val = list(val)[0]
                tmp_diff_vec = np.array(key)
                if (decoded_state + tmp_diff_vec[0])[val[0]] == val[1]:
                    diff_vec = tmp_diff_vec[0]
                    reward = tmp_diff_vec[1]
                    done = tmp_diff_vec[2]
                    break
                else:
                    reward = tmp_diff_vec[1]
                    done = tmp_diff_vec[2]
        else:
            diff_vec = np.array(optional_diff_vec[0][0])
            reward = optional_diff_vec[1]
            done = optional_diff_vec[2]
        not_legal_next_state = decoded_state + diff_vec
        state_is_legal, not_valid_idx = env.check_if_state_is_legal(not_legal_next_state, return_idxes=True)
        if not state_is_legal:
            new_not_legal_next_state = copy.deepcopy(not_legal_next_state)
            new_not_legal_next_state[not_valid_idx] = decoded_state[not_valid_idx]
            if (new_not_legal_next_state - decoded_state).any():
                mapping_states_dict[state] = (env.encode(*new_not_legal_next_state), reward, done)
            else:
                mapping_states_dict[state] = (state, reward, done)
        else:
            mapping_states_dict[state] = (env.encode(*not_legal_next_state), reward, done)
    return mapping_states_dict


def update_p_matrix_by_relax_preconditions(env, mapping_states_dict, pre_action, pre_prob_action=None):
    for state in mapping_states_dict:
        if pre_prob_action is not None:
            cur_info = list(env.P[state][pre_action][pre_prob_action])
            cur_info[1] = mapping_states_dict[state][0]
            cur_info[2] = mapping_states_dict[state][1]
            cur_info[3] = mapping_states_dict[state][2]
            env.P[state][pre_action][pre_prob_action] = cur_info
        else:  # deterministic case
            cur_info = list(env.P[state][pre_action][0])
            cur_info[1] = mapping_states_dict[state][0]
            cur_info[2] = mapping_states_dict[state][1]
            cur_info[3] = mapping_states_dict[state][2]
            env.P[state][pre_action] = [tuple(cur_info)]


def get_dicts(env, state, act_probs, act, next_state_dict_by_action, diff_dict_by_action, state_by_diff):
    next_state = act_probs[1]  # extract next state from the state prob
    reward = act_probs[2]
    done = act_probs[3]
    state_diff = np.array(env.decode(next_state)) - np.array(env.decode(state))
    if state_diff.any():
        next_state_dict_by_action[act].add((state, next_state))
        diff_dict_by_action[act].append([state_diff, reward, done])
        state_by_diff[act].add(state)
    return next_state_dict_by_action, diff_dict_by_action, state_by_diff


def get_diff_for_actions(p, env):
    next_state_by_action = [set() for act in range(len(p[0]))]
    diff_by_action = [list() for act in range(len(p[0]))]
    state_by_diff = [set() for act in range(len(p[0]))]
    for state, state_info in p.items():
        for act, act_probs in state_info.items():
            if len(act_probs) == 1:  # deterministic case
                next_state_by_action, diff_by_action, state_by_diff = get_dicts(env, state, act_probs[0], act,
                                                                                next_state_by_action, diff_by_action,
                                                                                state_by_diff)
            else:
                for act_prob, act_prob_info in enumerate(act_probs):
                    if len(act_prob_info) > 0:
                        next_state_by_action, diff_by_action, state_by_diff = get_dicts(env, state,
                                                                                        act_prob_info, act_prob,
                                                                                        next_state_by_action,
                                                                                        diff_by_action, state_by_diff)
    diff_dict = {}
    for i, action_diff in enumerate(diff_by_action):
        tmp_diff = [tuple(diff[0]) for diff in action_diff]
        occurrences = collections.Counter(tuple(tmp_diff))
        occurrences = occurrences.keys()
        reward = action_diff[0][1]
        done = action_diff[0][2] if (i != 5 and reward != 100) else True  # TODO - TEMPORARY! to change
        if len(occurrences) > 1:
            tmp_occ = list(occurrences)
            diff_dict[i] = dict(((x, reward, done), set()) for x in tmp_occ)
            for s, next_state in next_state_by_action[i]:
                for diff in tmp_occ:
                    decoded_state, decoded_next_state = np.array(env.decode(s)), np.array(env.decode(next_state))
                    state_diff = decoded_next_state - decoded_state
                    if not (state_diff - diff).any():
                        idx = np.nonzero(diff)
                        val = decoded_next_state[idx]
                        diff_dict[i][(diff, reward, done)].add((tuple(idx[0]), tuple(val)))
        else:
            diff_dict[i] = (list(occurrences), reward, done)
    return diff_dict, state_by_diff, next_state_by_action


def get_most_likely_outcome(p):
    for (s, s_probs) in p.items():
        for (a, a_probs) in s_probs.items():
            probs_list = [prob[0] for prob in a_probs]
            max_prob = max(probs_list)
            max_prob_idx = probs_list.index(max_prob)
            p[s][a] = [tuple([1.0] + list(p[s][a][max_prob_idx])[1:])]
    return p






from src.rlpe.Transforms.single_taxi_transforms import *


def generate_specific_transformed_env():
    #     env_default_values = [0, 0, 0, 1, MAX_FUEL - 1]
    # a_file = open("taxi_example_data/small_taxi_env_preconditions.pkl", "rb")
    # cur_env_preconditions = pickle.load(a_file)

    act1, act2, act3, act4, act5 = 0, 1, 2, 4, 5
    pre_idx = (4,)
    pre_val = [0]
    dir_name = TRANSFORMS_PATH
    env_file_name = f"{act1}_{pre_idx}_{pre_val}_{act2}_{pre_idx}_{pre_val}_{act3}_{pre_idx}_{pre_val}_{act4}_{pre_idx}_{pre_val}_{act5}_{pre_idx}_{pre_val}.pkl"
    precondition = {act1: {pre_idx: pre_val},
                    act2: {pre_idx: pre_val},
                    act3: {pre_idx: pre_val},
                    act4: {pre_idx: pre_val},
                    act5: {pre_idx: pre_val},
                    }
    env = generate_transformed_env(precondition, env_file_name=dir_name + env_file_name, save=True)

    print("DONE!")
    return env


def create_data_structure_with_all_envs():
    data_path = TRAINED_AGENT_ON_SPECIFIC_TRANSFORM_NUM_PATH
    all_single_transformed_envs = {}
    for file_name in os.listdir(data_path):
        transform_name = file_name[:-4]
        file = open(data_path + file_name, "rb")
        new_env = pickle.load(file)
        all_single_transformed_envs[transform_name] = new_env
    return all_single_transformed_envs


if __name__ == '__main__':
    #     all_single_transformed_envs = create_data_structure_with_all_envs()
    transformed_env = generate_specific_transformed_env()
    a = 7



SOUTH, NORTH, EAST, WEST, PICKUP, DROPOFF, REFUEL = 0, 1, 2, 3, 4, 5, 6
# PASSENGER_IN_TAXI = -1
STEP_REWARD, PICKUP_REWARD, BAD_PICKUP_REWARD, DROPOFF_REWARD, BAD_DROPOFF_REWARD, REFUEL_REWARD, BAD_REFUEL_REWARD, NO_FUEL_REWARD = "step", "good_pickup", "bad_pickup", "good_dropoff", "bad_dropoff", "good_refuel", "bad_refuel", "no_fuel"
ACTIONS = [SOUTH, NORTH, EAST, WEST, PICKUP, DROPOFF, REFUEL]
MAX_FUEL = 50
REWARD_DICT = {STEP_REWARD: -1,
               PICKUP_REWARD: 0, BAD_PICKUP_REWARD: -10,
               DROPOFF_REWARD: 100, BAD_DROPOFF_REWARD: -10,
               REFUEL_REWARD: 0, BAD_REFUEL_REWARD: -10, NO_FUEL_REWARD: -100}

DETERMINISTIC_PROB = 1.0
STOCHASTIC_PROB = 0.91
STOCHASTIC_PROB_OTHER_ACTIONS = (DETERMINISTIC_PROB - STOCHASTIC_PROB) / 3



class bidict(dict):
    def __init__(self, *args, **kwargs):
        super(bidict, self).__init__(*args, **kwargs)
        self.inverse = {}
        for key, value in self.items():
            self.inverse.setdefault(value, []).append(key)

    def __setitem__(self, key, value):
        if key in self:
            self.inverse[self[key]].remove(key)
        super(bidict, self).__setitem__(key, value)
        self.inverse.setdefault(value, []).append(key)

    def __delitem__(self, key):
        self.inverse.setdefault(self[key], []).remove(key)
        if self[key] in self.inverse and not self.inverse[self[key]]:
            del self.inverse[self[key]]
        super(bidict, self).__delitem__(key)



TAXI_ENVIRONMENT_REWARDS = dict(
    step=-1,
    no_fuel=-20,
    bad_pickup=-15,
    bad_dropoff=-15,
    bad_refuel=-10,
    bad_fuel=-50,
    pickup=50,
    standby_engine_off=-1,
    turn_engine_on=-10e6,
    turn_engine_off=-10e6,
    standby_engine_on=-1,
    intermediate_dropoff=-50,
    final_dropoff=100,
    hit_wall=-2,
    collision=-35,
    collided=-20,
    unrelated_action=-15
)

ALL_ACTIONS_NAMES = ['south', 'north', 'east', 'west',
                     'pickup', 'dropoff', 'refuel']
# ALL_ACTIONS_NAMES = ['south', 'north', 'east', 'west',
#                      'pickup', 'dropoff', 'bind',
#                      'turn_engine_on', 'turn_engine_off',
#                      'standby',
#                      'refuel']

BASE_AVAILABLE_ACTIONS = ['south', 'north', 'east', 'west',
                          'pickup', 'dropoff']



from src.rlpe.Observer.abstract_expert import AbstractExpert
from src.rlpe.Observer.taxi_expert import shortest_path_tree


class SingleTaxiExpert(AbstractExpert):
    def __init__(self, env):
        self.env = env
        self.desc = env.desc  # array holding map layout
        self.locs = env.passengers_locations  # locations of passenger pickup or dropoff X's on map
        self.num_rows = len(self.desc) - 2
        self.num_cols = len(self.desc[0][1:-1:2])
        self.max_row = self.num_rows - 1
        self.max_col = self.num_cols - 1
        # self.shortest_path_trees = []
        # for loc in self.locs:
        #     self.shortest_path_trees.append(shortest_path_tree(self.desc, loc))
        self.shortest_path_trees = {}
        for loc in env.passengers_locations:
            self.shortest_path_trees[tuple(loc)] = shortest_path_tree(self.desc, loc)
        self.ignored_taxi_locations = [(0, 1), (1, 1), (0, 2), (1, 2), (0, 3), (1, 3), (0, 4), (1, 4)]

    # get expert action, given a tuple of the form:
    # (taxi_loc_x, taxi_loc_y, fuel_level, pass_start_x, pass_start_y, pass_dest_x, pass_dest_y, pass_status)
    # unnecessary information in the tuple can be left as None
    # (e.g. fuel level is not used in this function and can be None)
    def get_expert_policy_set(self, state_tuple):
        (taxi_loc_x, taxi_loc_y, pass_loc_idx, pass_dest_idx, fuel) = state_tuple
        taxi_loc = (taxi_loc_x, taxi_loc_y)
        pass_loc = self.env.passengers_locations[pass_loc_idx] if pass_loc_idx != self.env.passenger_in_taxi else None
        pass_dest = self.env.passengers_locations[pass_dest_idx]
        if pass_loc_idx == self.env.passenger_in_taxi and taxi_loc == pass_dest:
            return [DROPOFF]  # dropoff
        elif pass_loc_idx == self.env.passenger_in_taxi and taxi_loc != pass_dest:
            return self.shortest_path_trees[pass_dest][taxi_loc_x][taxi_loc_y][0]
        elif pass_loc_idx != self.env.passenger_in_taxi and taxi_loc == pass_loc and pass_loc_idx != pass_dest_idx:
            return [PICKUP]  # pickup
        elif pass_loc_idx != self.env.passenger_in_taxi and taxi_loc != pass_loc and pass_loc_idx != pass_dest_idx:
            return self.shortest_path_trees[pass_loc][taxi_loc_x][taxi_loc_y][0]
        elif pass_loc_idx != self.env.passenger_in_taxi and pass_loc_idx == pass_dest_idx:
            # print("Passenger has already arrived")
            return None
        else:
            print("Not a valid state")

    def full_expert_policy_dict(self):
        fuel_level = None
        expert_policy_dict = {}
        for taxi_loc_x in range(self.num_rows):
            for taxi_loc_y in range(self.num_cols):
                if (taxi_loc_x, taxi_loc_y) in self.ignored_taxi_locations:
                    continue
                for pass_start_idx in range(len(self.env.passengers_locations)):
                    for pass_dest_idx in range(len(self.env.passengers_locations)):
                        state_tuple = (taxi_loc_x, taxi_loc_y, pass_start_idx, pass_dest_idx, fuel_level)
                        optimal_action = self.get_expert_policy_set(state_tuple)
                        if optimal_action:
                            expert_policy_dict[state_tuple] = optimal_action
        return expert_policy_dict



from abc import ABC, abstractmethod


class AbstractWrapperEnv(ABC):
    """
    This is the basic AbstractTransformedEnvironment requirements for working with this project.
    """

    @abstractmethod
    def get_states_from_partial_obs(self, partial_obs):
        """
        ...
        """
        pass

    @abstractmethod
    def encode(self, state):
        """
        ...
        """
        pass

    @abstractmethod
    def decode(self, i):
        pass




from src.rlpe.Observer.abstract_expert import AbstractExpert
import numpy as np
import json

PARTIAL_POLICY_IDX = 0  # Should be 0 <= i <=5


class LunarLanderExpert(AbstractExpert):
    def __init__(self, env):
        self.env = env

    def get_expert_policy_set(self, state_tuple):
        pass

    def full_expert_policy_dict(self):
        prob_dict_tmp = {}
        prob_dict = {}
        full_expert_policy_dict = {}
        expert_policy_dict = {}

        with open('results/sarsa_data/sarsa_Q_5X4Ys.json') as json_file:
            Q = json.load(json_file)
            for k, v in Q.items():
                state_name = tuple([int(x) for x in (k[1:-3].split(','))])
                if state_name not in prob_dict_tmp.keys():
                    prob_dict_tmp[state_name] = [0] * 4
                prob_dict_tmp[state_name][int(k[-1])] = v
        for k, val in prob_dict_tmp.items():
            max_val = max(val)
            for i in range(len(val)):
                if val[i] == max_val:
                    prob_dict[k] = i
            full_expert_policy_dict[k] = self.heuristic(list(k))
        for key, value in full_expert_policy_dict.items():
            key = list(key)
            state = tuple(key[:PARTIAL_POLICY_IDX] + [None] + key[PARTIAL_POLICY_IDX + 1:])
            if state not in expert_policy_dict.keys():
                expert_policy_dict[state] = value

        return expert_policy_dict

    def heuristic(self, s):
        # Heuristic for:
        # 1. Testing.
        # 2. Demonstration rollout.

        # angle should point towards center (s[0] is horizontal coordinate, s[2] hor speed)
        angle_targ = s[0] * 0.5 + s[2] * 1.0
        if angle_targ > 0.4:
            angle_targ = 0.4  # more than 0.4 radians (22 degrees) is bad
        if angle_targ < -0.4:
            angle_targ = -0.4
        hover_targ = 0.55 * np.abs(s[0])  # target y should be proportional to horizontal offset

        # PID controller: s[4] angle, s[5] angularSpeed
        angle_todo = (angle_targ - s[4]) * 0.5 - (s[5]) * 1.0
        # print("angle_targ=%0.2f, angle_todo=%0.2f" % (angle_targ, angle_todo))

        # PID controller: s[1] vertical coordinate s[3] vertical speed
        hover_todo = (hover_targ - s[1]) * 0.5 - (s[3]) * 0.5
        # print("hover_targ=%0.2f, hover_todo=%0.2f" % (hover_targ, hover_todo))

        if s[6] or s[7]:  # legs have contact
            angle_todo = 0
            hover_todo = -(s[3]) * 0.5  # override to reduce fall speed, that's all we need after contact

        if self.env.continuous:
            a = np.array([hover_todo * 20 - 1, -angle_todo * 20])
            a = np.clip(a, -1, +1)
        else:
            a = 0
            if hover_todo > np.abs(angle_todo) and hover_todo > 0.05:
                a = 2
            elif angle_todo < -0.05:
                a = 3
            elif angle_todo > +0.05:
                a = 1
        return a

    def demo_heuristic_lander(self, env, seed=None, render=False):
        env.seed(seed)
        total_reward = 0
        steps = 0
        s = env.reset()
        while True:
            a = self.heuristic(s)
            s, r, done, info = env.step(a)
            total_reward += r

            if render:
                still_open = env.render()
                if not still_open:
                    break

            if (steps % 20 == 0 or done) and render:
                print("observations:", " ".join(["{:+0.2f}".format(x) for x in s]))
                print("step {} total_reward {:+0.2f}".format(steps, total_reward))
            steps += 1
            if done:
                break
        return total_reward
